
;;;*************************************************************************
;;;module name: cl-semprop1.l
;;;
;;;author     : Reinhold Letz, TU Munich
;;;
;;;date       : 09.02.94
;;;modified   :	
;;;
;;;description: propositional decision procedure using the semantic tree 
;;;		calculus incorporating different selection functions 
;;;		and lemmata generated at proper branching points
;;;		with marking of the used clauses in the refutation;
;;;		compiler is included
;;;
;;;remarks    : literals may occur multiply in a clause, also tautological
;;;		clauses may occur
;;;		all global program variables are prefixed by a `$',
;;;		therefore no atom in the input should have `$' as a prefix
;;;		as in cl-semprop.l but data structures atom, clause, and
;;;		folded replaced by vectors
;;;compiling  : yes
;;;*************************************************************************

;; -------------------------------------------------------------------------
;; internal data structures:
;; 
;; clause: vector of seven elements
;; 1st element: atom which activates the clause [nil or atom]
;; 2nd element: current clause length [positive integer]
;; 3rd element: list of positive atoms
;; 4th element: list of negative atoms
;; 5th element: initial clause length [positive integer]
;; 6th element: number of clause usages [natural number]
;; 7th element: clause number [positive integer]
;;
;; folded: vector of three elements
;; 1st element: folded atom
;; 2nd element: list of positive context atoms
;; 3rd element: list of negative context atoms
;;
;; atom: vector of sixteen elements
;; 1st element: atom is currently unevaluated [nil] or evaluated ['l or 'r] 
;; 2nd element: clauses with positive occurrence of atom
;; 3rd element: clauses with negative occurrence of atom
;; 4th element: last inference using assignment of atom
;; 5th element: positive atom frequency in reduced formula
;; 6th element: negative atom frequency in reduced formula
;; 7th element: empty clause in context [t] or not [nil]
;; 8th element: list of context unit clauses with positive literals
;; 9th element: list of non-unit clauses to be inserted
;;10th element: atom name [symbol]
;;11th element: last cdr of 2nd element
;;12th element: last cdr of 3rd element
;;13th element: list of non-unit clauses to be deleted at this point
;;14th element: list of context unit clauses with negative literals
;;15th element: pos_counter for clause generation from ground enumerator
;;16th element: neg_counter for clause generation from ground enumerator
;; 
;; global variables:
;;
;; $clauselist: list of clauses to be refuted
;; $atomlist: list of atoms in formula
;; $units: list of unit clauses 
;; -------------------------------------------------------------------------

;; -------------------------------------------------------------------------
;; compilation instructions:
(proclaim '(optimize (speed 3) (compilation-speed 0)
                     (space 0) (safety 0) (debug 0)))
;; -------------------------------------------------------------------------

;; -------------------------------------------------------------------------
;; macros for accessing basic data structures
;;
(defmacro c-activating-atom 	(clause) `(svref ,clause 0))
(defmacro c-length 		(clause) `(svref ,clause 1))
(defmacro c-pos-literals 	(clause) `(svref ,clause 2))
(defmacro c-neg-literals 	(clause) `(svref ,clause 3))
(defmacro c-init-length		(clause) `(svref ,clause 4))
(defmacro c-usages		(clause) `(svref ,clause 5))
(defmacro c-number		(clause) `(svref ,clause 6))

(defmacro f-atom		(folded) `(svref ,folded 0))
(defmacro f-pos-atoms		(folded) `(svref ,folded 1))
(defmacro f-neg-atoms		(folded) `(svref ,folded 2))

(defmacro a-evaluated		(atom) `(svref ,atom 0))
(defmacro a-pos-clauses		(atom) `(svref ,atom 1))
(defmacro a-neg-clauses 	(atom) `(svref ,atom 2))
(defmacro a-last-inference	(atom) `(svref ,atom 3))
(defmacro a-pos-frequency	(atom) `(svref ,atom 4))
(defmacro a-neg-frequency	(atom) `(svref ,atom 5))
(defmacro a-falsum		(atom) `(svref ,atom 6))
(defmacro a-pos-folded          (atom) `(svref ,atom 7))
(defmacro a-lemmata             (atom) `(svref ,atom 8))
(defmacro a-name                (atom) `(svref ,atom 9))
(defmacro a-last-pos-clauses    (atom) `(svref ,atom 10))
(defmacro a-last-neg-clauses    (atom) `(svref ,atom 11))
(defmacro a-delete-lemmata      (atom) `(svref ,atom 12))
(defmacro a-neg-folded          (atom) `(svref ,atom 13))
(defmacro a-pos-counter		(atom) `(svref ,atom 14))
(defmacro a-neg-counter		(atom) `(svref ,atom 15))

;; -------------------------------------------------------------------------
;; numbers encoding different selection functions
;;
;; $selection = 0: given atom ordering
;; $selection = 1: atoms sorted initially
;; $selection = 2: dynamic selection of the respective maximal atom 
;; $selection = 3: tableaux selection, else given atom ordering
;; $selection = 4: tableaux selection, else initially sorted ordering
;; $selection = 5: tableaux selection, else dynamical selection of maximal atom 
;; $selection = 6: maximal literal tableaux selection, 
;;		   else initially sorted ordering
;; $selection = 7: maximal literal tableaux selection from shortest clause, 
;;		   else initially sorted ordering
;; $selection = 8: random selection
;; $selection = 9: selection of atom in shortest clause
;;
;; -------------------------------------------------------------------------
;; default options: 
(setq $selection 1)	;; atoms sorted statically in advance
(setq $dynamic nil)	;; no dynamic atom selection is performed 
(setq $verbose 0)	;; if a model is found it is printed
;$verbose = 1:		;; if a model is found it is not printed, but stored 
			;; as first element in the global variable $models 
;$verbose = anything else: ;; in case of a model, only nil is returned 
;$output = positive int n: ;; n literals are printed in a line (default = 10)
(setq $max-lemma-store 10000)  ;; 10000 lemmata kept
(setq $lemmata 2)	;; permanent lemmata are dynamically generated
;$lemmata = 1:	 	;; folding up (temporary context-unit lemmata)
;$lemmata = 0:	 	;; no folding up, no permanent lemmata
(setq $lemma-h 30)	;; generates lemmata only if subproof >= 1 step
;(setq $int-length 20)	;; generates lemmata only if path length <= 20
(setq $clever-insert nil);; lemmata are inserted and deleted in fifo-style
;$clever-insert = t 	;; lemmata are inserted and deleted according to 
			;; their lengths
(setq $deletion nil)	;; lemmas are not deleted
(defun verbose-prove ()
(let (result)
(terpri)
(setq result (prove))
(cond (result (terpri) (princ "Formula refuted after ") (prin1 result)
              (princ " inferences") (terpri) 
        ;      (princ "The following clause set is unsatisfiable:") (terpri)
      )
      (t 'Satisfiable)
)))

(defun prove () (declare (fixnum $selection) (special $selection))
(setq $atom-nr (length $atomlist))
(setq $lemmalist nil)	;; list of lemmata initially empty
(setq $lemma-store 0) 	;; initial size of lemma store 
(setq $model nil) 	;; no model found up to now
(setq $dynamic nil)
; (setq $fc 0)
; (setq $del 0)
 (init-inf)
 (cond ((or (= $selection 2) (= $selection 5)) (setq $dynamic t)))
 (cond ((or (= $selection 1) (= $selection 4) (> $selection 5))
        (order-atoms $atomlist)))
 (falsify-left nil $units nil nil $atomlist 0))

;; -------------------------------------------------------------------------
;; main procedure: performs specialization of (partial) interpretations
;; initial call: (falsify-left nil $units nil nil $atomlist 0)

(defun falsify-left ;; performs left branching in semantic tree
  (interpretation unit-clauses p-folded n-folded atoms inference)
(declare (special $lemma-h $lemmata $lemma-tag $max-lemma-store $selection)
       (fixnum $lemma-h $lemmata $max-lemma-store $selection sign inference))
(let (folded atom sign)
(declare (fixnum sign))
 (cond
;; first it is checked whether a positive literal L has been folded up
  ((setq p-folded (find-active-folded p-folded))
   (setq folded (car p-folded)) (setq sign 1) (setq atom (f-atom folded))
   (setq inference (1+ inference)) (rplcatinf folded inference)
   (cond
;;;; if yes, then it is checked whether ~L has been folded up, too
     ((setq folded (is-fold-literal atom n-folded))
;;;;;; if yes, then the partial interpretation can be immediately falsified
      (setq inference (1+ inference)) (rplcatinf folded inference) inference)
;;;;;; if no, it is proceeded in right branching mode
     (t (left-unit-right-evaluate interpretation unit-clauses p-folded
                                  n-folded atoms atom sign inference))))
;; if no positive folded literal is available, it is checked whether a
;; negative literal has been folded up
  ((setq n-folded (find-active-folded n-folded))
   (setq folded (car n-folded)) (setq sign -1) (setq atom (f-atom folded))
   (setq inference (1+ inference)) (rplcatinf folded inference)
;;;; afterwards, it is proceeded in right branching mode, as above
   (left-unit-right-evaluate interpretation unit-clauses nil n-folded
                             atoms atom sign inference))
;; if neither positive nor negative folded literals are available, then
;; it is checked whether an active clause has current length 1,
  ((setq unit-clauses (find-active unit-clauses))
;; in which case the respective literal L from the left-most one is selected
  (setq inference (1+ inference))
   (get-literal-in-unit-clause (car unit-clauses) atom sign inference)
;; -------------------------------------------------------------------------
;; afterwards the current partial interpretation is extended by the literal
;; L and the algorithm proceeds by evaluating the right branch
   (left-unit-right-evaluate 
         interpretation unit-clauses nil nil atoms atom sign inference))
;; -------------------------------------------------------------------------
  ((setq atoms (rem-prefix atoms)) 
;; if no falsified unit clause can be found and there are still atoms 
;; available, a proper branching has to occur
   (branching interpretation atoms atom sign inference))
;; -------------------------------------------------------------------------
  (t (setq $model t) (setq $end-inference inference)
     (model interpretation inference)))))
;; otherwise all atoms have been evaluated and a model has been found
;; -------------------------------------------------------------------------

(defmacro rplcatinf (folded inference)
`(progn (rplac-atom-inference (f-pos-atoms ,folded) ,inference)
        (rplac-atom-inference (f-neg-atoms ,folded) ,inference)))

(defun find-active-folded (foldeds)
 (cond (foldeds
        (cond 
         ((a-evaluated (f-atom (car foldeds))) 
          (find-active-folded (cdr foldeds)))
         (t foldeds)))))

(defun is-fold-literal (atom foldeds)
 (cond (foldeds
        (cond ((eq atom (f-atom (car foldeds))) (car foldeds))
              (t (is-fold-literal atom (cdr foldeds)))))))

(defun find-active (clauses)
 (cond (clauses
        (cond ((c-activating-atom (car clauses)) (find-active (cdr clauses)))
              (t clauses)))))

(defmacro get-literal-in-unit-clause (clause atom sign inference)
 `(progn
   (or (and (setq ,atom (get-literal (c-pos-literals ,clause)))
            (setq ,sign 1)) ;; negative valuation of atom checked off
       (and (setq ,atom (get-literal (c-neg-literals ,clause)))
            (setq ,sign -1)));; positive valuation of atom checked off
   (setf (c-usages ,clause) (1+ (c-usages ,clause)))
   (rplac-atom-inference (c-pos-literals ,clause) ,inference)
   (rplac-atom-inference (c-neg-literals ,clause) ,inference)
))

(defmacro left-unit-right-evaluate ;; evaluates a right branch after the left
                                   ;; one could be falsified by a unit clause
 (interpretation unit-clauses p-folded n-folded atoms atom sign inference)
; (declare (fixnum inference sign))
 `(let (output clauses stop-clause true-clauses new-units new-inference)
   (declare (type fixnum new-inference))
   (cond ((= ,sign -1) ;; negative valuation of atom to be checked
          (setq clauses (a-pos-clauses ,atom))
          (setq true-clauses (a-neg-clauses ,atom)))
         (t ;; positive valuation of atom to be checked
          (setq clauses (a-neg-clauses ,atom))
          (setq true-clauses (a-pos-clauses ,atom))))
   (setq output (decrease-counters clauses nil))
   (cond 
    ((setq stop-clause (car output)) ;; a stop-clause exists
;; existence of a stop clause means that unit clauses for an atom and its
;; negation exist, in which case the application of unit clauses is stopped 
;; and the whole branch is closed
     (increase-counters clauses stop-clause)
     (setq ,inference (1+ ,inference))
     (setf (c-usages stop-clause) (1+ (c-usages stop-clause)))
     (rplac-atom-inference (c-pos-literals stop-clause) ,inference)
     (rplac-atom-inference (c-neg-literals stop-clause) ,inference)
     ,inference)
    (t (setq new-units (cdr output))
;; -------------------------------------------------------------------------
;; if no stop clause exists, all unit clauses have been applied and a right
;; branch cannot be immediately extended by an atom with the resulting partial 
;; interpretation falsifying a unit clause
     (disactivate true-clauses ,atom)
     (setf (a-evaluated ,atom) 'r);; atom is made unavailable in right mode
     (setq new-inference
           (falsify-left (cons (cons ,sign ,atom) ,interpretation) 
                         (nconc new-units ,unit-clauses) ,p-folded ,n-folded
                         ,atoms ,inference))
     (setf (a-evaluated ,atom) nil);; atom is made available
     (remove-context-clauses ,atom);; context lemmata are deleted
     (reactivate true-clauses ,atom) ;; clauses are reactivated
     (incr-counters clauses) ;; clause lengths are incremented
     new-inference)
)))

(defmacro branching ;; a proper semantic tree branching is performed
                    (interpretation atoms atom sign inference)
 `(let (clauses true-clauses new-units new-inference1 new-inference2)
   (declare (fixnum new-inference1 new-inference2))
;; first, the atom selection is performed according to the specified parameter
   (cond ((< $selection 2) (setq ,atom (car ,atoms)))
         ((= $selection 2) (setq ,atom (d-max-atom ,atoms)))
         ((> 7 $selection 2) 
          (cond ((setq ,atom (select-atom ,interpretation)) nil)
                ((= $selection 5) (setq ,atom (d-max-atom ,atoms)))
                (t (setq ,atom (car ,atoms)))))
         ((= $selection 7)
          (let (short-clause)
          (cond
           ((setq short-clause (select-atom-shortest ,interpretation))
            (let (max pos neg)
             (setq pos (rem-prefix (c-pos-literals short-clause)))
             (setq neg (rem-prefix (c-neg-literals short-clause)))
             (cond (pos (setq max (d-max-atom pos))
                    (cond (neg (setq neg (cons max neg)) 
                           (setq ,atom (d-max-atom neg)))
                          (t (setq ,atom max))))
                   (t (setq ,atom (d-max-atom neg))))))
           (t (setq ,atom (car ,atoms))))))
         ((= $selection 8)
          (setq ,atom (random-atom ,atoms)))
         ((= $selection 9)
          (let (short-clause)
          (cond
           ((setq short-clause (select-shortest $lemmalist $clauselist nil 0))
            (let (max pos neg)
             (setq pos (rem-prefix (c-pos-literals short-clause)))
             (setq neg (rem-prefix (c-neg-literals short-clause)))
             (cond (pos (setq max (d-max-atom pos))
                    (cond (neg (setq neg (cons max neg)) 
                           (setq ,atom (d-max-atom neg)))
                          (t (setq ,atom max))))
                   (t (setq ,atom (d-max-atom neg))))))
           (t (setq ,atom (car ,atoms))))))
   )
;; -------------------------------------------------------------------------
   (setf (a-evaluated ,atom) 'l) ;; atom is made unavailable in left mode
   (cond ((< (a-pos-frequency ,atom) (a-neg-frequency ,atom))
          (setq ,sign -1) (setq clauses (a-pos-clauses ,atom))
          (setq true-clauses (a-neg-clauses ,atom)))
         (t (setq ,sign 1) (setq clauses (a-neg-clauses ,atom))
            (setq true-clauses (a-pos-clauses ,atom))))
   (setq new-units (decr-counters clauses nil))
   (disactivate true-clauses ,atom)
   (setq new-inference1 
         (falsify-left (cons (cons ,sign ,atom) ,interpretation)
         new-units nil nil ,atoms ,inference))
   (reactivate true-clauses ,atom) ;; clauses are reactivated
   (incr-counters clauses) ;; clause lengths are incremented
   (cond 
    (new-inference1 
     (cond ((= 2 $lemmata)
            (setq $atom ,atom) 
            (cond ($clever-insert (replace-clauses (a-lemmata ,atom)))
                  (t (attach-clauses (a-lemmata ,atom))))
            (delete-clauses (a-delete-lemmata ,atom))))
     (cond 
      ((or (a-falsum ,atom) (<= (a-last-inference ,atom) ,inference))
       (setf (a-evaluated ,atom) nil);; atom is made available
       (remove-context-clauses ,atom);; context lemmata are deleted
       new-inference1)
      (t 
       (setq new-units (decr-counters true-clauses nil))
       (setf (a-evaluated ,atom) 'r) ;; atom is made unavailable in right mode
       (disactivate clauses ,atom)
       (cond ((= ,sign 1) (setq ,sign -1)) (t (setq ,sign 1)))
       (setq new-inference2
             (falsify-left (cons (cons ,sign ,atom) ,interpretation) 
                           new-units (a-pos-folded ,atom) (a-neg-folded ,atom) 
                           ,atoms new-inference1))
       (setf (a-evaluated ,atom) nil);; atom is made available
       (reactivate clauses ,atom) ;; clauses are reactivated
       (incr-counters true-clauses) ;; clause lengths are incremented
       (remove-context-clauses ,atom);; context lemmata are deleted
;       (cond ((> (length ,interpretation) $int-length) nil)
;             (t (insert-clause0 ,interpretation ,inference nil)))
;----------------------------------------------------------------------------
       (cond ((and (null $model) (> $lemmata 0)) 
              (cond ((<= $lemma-h (- new-inference2 ,inference))
                     (setq $lemma-tag t))
                    (t (setq $lemma-tag nil)))
              (insert-clause-iterative ,interpretation ,inference)))
;----------------------------------------------------------------------------
       new-inference2)
    ))
    (t (setf (a-evaluated ,atom) nil);; atom is made available
       (remove-context-clauses ,atom);; context lemmata are deleted
       nil)
)))

(defmacro remove-context-clauses (atom)
 `(progn (setf (a-falsum ,atom) nil)
         (setf (a-pos-folded ,atom) nil)
         (setf (a-lemmata ,atom) nil)
         (setf (a-delete-lemmata ,atom) nil)
         (setf (a-neg-folded ,atom) nil)))

(defun decrease-counters ;; decreases counters with considering stop clauses
                         (clauses new-units)
;; output is a dotted pair: (a . b)
;; a is the stop clause if existing or nil
;; b is the list of new-units resulting from the specialization if a is nil
 (cond 
  (clauses 
   (let (clause) (setq clause (car clauses))
        (cond ((c-activating-atom clause) ;; clause is not active
               (decrease-counters (cdr clauses) new-units))
              ((= 1 (c-length clause)) ;; complement falsifies a unit clause
               (cons clause nil))
              ((= 2 (c-length clause)) ;; a new unit clause is produced
               (setf (c-length clause) 1)
               (decrease-counters (cdr clauses) (cons clause new-units)))
              (t (setf (c-length clause) (1- (c-length clause)))
                 (decrease-counters (cdr clauses) new-units)))))
  (t (cons nil new-units))
))

(defun decr-counters ;; decreases counters without considering stop clauses
                     (clauses new-units) 
 (cond 
  (clauses 
   (let (clause) (setq clause (car clauses))
        (cond ((c-activating-atom clause) ;; clause is not active
               (decr-counters (cdr clauses) new-units))
              ((= 2 (c-length clause)) ;; a new unit clause is produced
               (setf (c-length clause) 1)
               (decr-counters (cdr clauses) (cons clause new-units)))
              (t (setf (c-length clause) (1- (c-length clause)))
                 (decr-counters (cdr clauses) new-units)))))
  (t new-units)
))

(defun increase-counters ;; increases counters with considering stop clause
                         (clauses stop-clause)
 (cond 
  (clauses 
   (let (clause) (setq clause (car clauses))
        (cond ((eq stop-clause clause) nil)
              ((c-activating-atom clause) ;; clause is not active
               (increase-counters (cdr clauses) stop-clause))
              (t (setf (c-length clause) (1+ (c-length clause)))
                 (increase-counters (cdr clauses) stop-clause))
)))))

(defun incr-counters ;; increases counters without considering stop clause
                     (clauses)
 (cond 
  (clauses 
   (let (clause) (setq clause (car clauses))
        (cond ((null (c-activating-atom clause)) ;; clause was active
               (setf (c-length clause) (1+ (c-length clause)))))
        (incr-counters (cdr clauses))))))

(defun disactivate (clauses atom)
 (declare (special $dynamic))
 (cond 
  (clauses 
   (let (clause) (setq clause (car clauses))
        (cond ((null (c-activating-atom clause)) ;; clause is active
               (setf (c-activating-atom clause) atom) 
               (cond ($dynamic (decrement-frequencies clause)))
        ))
        (disactivate (cdr clauses) atom)))))

(defmacro decrement-frequencies (clause)
 `(progn (decr-pos-freq (c-pos-literals ,clause))
         (decr-neg-freq (c-pos-literals ,clause))))

(defun decr-pos-freq (atoms)
 (cond 
  (atoms 
   (let (atom) (setq atom (car atoms))
        (setf (a-pos-frequency atom) (1- (a-pos-frequency atom)))
        (decr-pos-freq (cdr atoms))))))

(defun decr-neg-freq (atoms)
 (cond 
  (atoms 
   (let (atom) (setq atom (car atoms))
        (setf (a-neg-frequency atom) (1- (a-neg-frequency atom)))
        (decr-neg-freq (cdr atoms))))))

(defun reactivate (clauses atom)
 (declare (special $dynamic))
 (cond 
  (clauses 
   (let (clause) (setq clause (car clauses))
        (cond ((eq (c-activating-atom clause) atom) 
               (setf (c-activating-atom clause) nil) ;; clause is reactivated
               (cond ($dynamic (increment-frequencies clause)))
        ))
        (reactivate (cdr clauses) atom)))))

(defmacro increment-frequencies (clause)
 `(progn (incr-pos-freq (c-pos-literals ,clause))
         (incr-neg-freq (c-pos-literals ,clause))))

(defun incr-pos-freq (atoms)
 (cond 
  (atoms 
   (let (atom) (setq atom (car atoms))
        (setf (a-pos-frequency atom) (1+ (a-pos-frequency atom)))
        (incr-pos-freq (cdr atoms))))))

(defun incr-neg-freq (atoms)
 (cond 
  (atoms 
   (let (atom) (setq atom (car atoms))
        (setf (a-neg-frequency atom) (1+ (a-neg-frequency atom)))
        (incr-neg-freq (cdr atoms))))))

(defun rplac-atom-inference (atoms inference)
(declare (fixnum inference)) 
 (cond (atoms (setf (a-last-inference (car atoms)) inference)
              (rplac-atom-inference (cdr atoms) inference))
))

(defun get-literal (clause-half)
 (cond (clause-half
        (cond ((null (a-evaluated (car clause-half)))
               (car clause-half))
              (t (get-literal (cdr clause-half)))
))))

(defmacro select-atom-shortest (interpretation) 
 `(cond 
   (,interpretation 
    (let (clauses)
     (cond
      ((= (caar ,interpretation) 1)
       (setq clauses (a-neg-clauses (cdar ,interpretation))))
      (t (setq clauses (a-pos-clauses (cdar ,interpretation)))))
     (select-shortest nil clauses nil 0)))))

(defun rem-prefix (atoms)
 (cond (atoms 
        (cond ((a-evaluated (car atoms)) (rem-prefix (cdr atoms))) 
              (t atoms)))))

(defun select-shortest (lemmata clauses clause length)
 (declare (fixnum length))
  (setq lemmata (find-active lemmata))
  (cond
   ((null lemmata) 
    (cond
     ((setq clauses (find-active clauses))
      (setq clause (car clauses)) (setq length (c-length clause))
      (cond ((= 2 length) clause)
            (t (setq clause (select-shortest1 (cdr clauses) clause length))
               clause)))))
   (t (setq clause (car lemmata)) (setq length (c-length clause))
    (cond ((= 2 length) clause)
          (t (setq clause (select-shortest1 (cdr lemmata) clause length))
             (setq length (c-length clause))
           (cond
            ((= length 2) clause)
            (t (setq clause (select-shortest1 clauses clause length))
               clause)))))
))

(defun select-shortest1 (clauses old-clause old-length)
(declare (fixnum old-length))
(prog (clause length) 
(declare (fixnum length))
loop
 (cond 
  (clauses (setq clause (car clauses)) (setq length (c-length clause))
   (cond ((c-activating-atom clause) 
          (setq clauses (cdr clauses)) (go loop))
         (t 
          (cond ((= length 2) (return clause))
                ((< length old-length)
                 (setq old-clause clause) (setq old-length length)
                 (setq clauses (cdr clauses)) (go loop))
                (t (setq clauses (cdr clauses)) (go loop))))))
  (t (return old-clause))
)))

(defmacro short-select-neighbour (sign atom)
 (declare (fixnum sign length))
 `(let (clauses)
    (cond ((= ,sign 1) (setq clauses (a-neg-clauses ,atom)))
          (t (setq clauses (a-pos-clauses ,atom))))
    (short-select-neighbour-atom clauses)))

(defun short-select-neighbour-atom (clauses)
 (let (clause)
  (cond 
   (clauses (setq clause (car clauses))
    (cond ((c-activating-atom clause)
           (short-select-neighbour-atom (cdr clauses)))
          ((= 2 (c-length clause)) 
           (or (get-literal (c-pos-literals clause))
               (get-literal (c-neg-literals clause))))
          (t (short-select-neighbour-atom (cdr clauses)))
)))))
 
(defmacro select-atom (interpretation) 
 `(cond 
   (,interpretation 
    (cond
     ((= (caar ,interpretation) 1)
      (select-neighbour-atom (a-neg-clauses (cdar ,interpretation))))
     (t (select-neighbour-atom (a-pos-clauses (cdar ,interpretation))))))))

(defun select-neighbour-atom (clauses)
(declare (special $selection) (fixnum $selection))
 (let (clause)
  (cond 
   (clauses (setq clause (car clauses))
    (cond 
     ((c-activating-atom clause)
      (select-neighbour-atom (cdr clauses)))
     ((< $selection 6) 
      (or (get-literal (c-pos-literals clause))
          (get-literal (c-neg-literals clause))))
     (t (let (max pos neg)
         (setq pos (rem-prefix (c-pos-literals clause)))
         (setq neg (rem-prefix (c-neg-literals clause)))
         (cond (pos (setq max (d-max-atom pos))
                (cond (neg (setq neg (cons max neg)) (d-max-atom neg))
                      (t max)))
               (t (d-max-atom neg))))))))))
 
(defun order-atoms (atoms)
 (let (max-atom max-atom-pred)
 (cond (atoms (setq max-atom-pred (max-atom atoms))
              (setq max-atom (car max-atom-pred))
        (cond ((eq max-atom (car atoms)) nil)
              (t (rplaca max-atom-pred (car atoms))
                 (rplaca atoms max-atom)))
        (order-atoms (cdr atoms))))))

(defmacro max-atom (atoms)
 `(let (m-pos m-neg)
   (declare (fixnum m-pos m-neg))
   (setq m-pos (a-pos-frequency (car ,atoms))) 
   (setq m-neg (a-neg-frequency (car ,atoms)))
   (cond ((< m-pos m-neg) 
          (max-atom-atoms (cdr ,atoms) ,atoms m-neg m-pos))
         (t (max-atom-atoms (cdr ,atoms) ,atoms m-pos m-neg)))))

(defun max-atom-atoms (atoms atom-pred max min) 
 (declare (fixnum max min))
 (let (pos neg new-atom)
  (declare (fixnum pos neg))
 (cond 
  (atoms 
    (setq new-atom (car atoms))
    (setq pos (a-pos-frequency new-atom))
    (setq neg (a-neg-frequency new-atom))
    (cond 
     ((>= pos neg)
      (cond 
       ((> max pos) (max-atom-atoms (cdr atoms) atom-pred max min))
       ((= max pos)
        (cond ((>= min neg)
               (max-atom-atoms (cdr atoms) atom-pred max min))
              (t (max-atom-atoms (cdr atoms) atoms pos neg))))
       (t (max-atom-atoms (cdr atoms) atoms pos neg))))
     ;; in the following: neg > pos:
     ((> max neg) (max-atom-atoms (cdr atoms) atom-pred max min))
     ((= max neg)
      (cond ((>= min pos)
             (max-atom-atoms (cdr atoms) atom-pred max min))
              (t (max-atom-atoms (cdr atoms) atoms neg pos))))
     (t (max-atom-atoms (cdr atoms) atoms neg pos))))
  (t atom-pred))))

(defun insert-clause-iterative (interpretation inference)
(declare (special $deletion $lemmata $lemma-tag) (fixnum $lemmata))
(prog (atom fold-sign fold-atom length pos-lits neg-lits insertion 
       deletion clause)
(declare (type fixnum fold-sign length))
state0
(cond (interpretation (setq atom (cdar interpretation))
       (cond ((<= (a-last-inference atom) inference)
              (cond ((eq (a-evaluated atom) 'r) nil) 
                    ((a-falsum atom) nil)
                    (t (setf (a-falsum atom) t)))
              (setq interpretation (cdr interpretation)) (go state0))
             (t 
              (cond 
               ((= (caar interpretation) 1) (setq fold-sign -1))
               (t (setq fold-sign 1))) 
              (setq fold-atom (vector atom nil nil))
              (setq interpretation (cdr interpretation)) (go state1)))))
state1
(cond (interpretation (setq atom (cdar interpretation))
        (cond ((<= (a-last-inference atom) inference)
               (cond ((eq (a-evaluated atom) 'r) nil) 
                     ((a-falsum atom) nil)
                     ((= fold-sign 1) 
                      (setf (a-pos-folded atom) 
                            (cons fold-atom (a-pos-folded atom))))
                     (t (setf (a-neg-folded atom) 
                              (cons fold-atom (a-neg-folded atom)))))
               (setq interpretation (cdr interpretation)) (go state1))
              (t
               (cond
                ((= (caar interpretation) 1) (setq neg-lits (list atom)))
                (t (setq pos-lits (list atom))))
               (setq length 2)
               (setq interpretation (cdr interpretation)) (go state2))))
      (t (return t)))
state2
(cond (interpretation (setq atom (cdar interpretation))
       (cond ((<= (a-last-inference atom) inference)
              (cond ((or (eq (a-evaluated atom) 'r) (a-falsum atom)) 
                     (setq interpretation (cdr interpretation)) (go state2))
                    (t 
                     (cond 
                      ((and (= $lemmata 2) $lemma-tag)
                       (setq insertion atom)))
                     (setq interpretation (cdr interpretation)) (go state3))))
             (t (setq length (1+ length))
              (cond 
               ((= (caar interpretation) 1)
                (setq neg-lits (cons atom neg-lits)))
               (t (setq pos-lits (cons atom pos-lits))))
              (setq interpretation (cdr interpretation)) (go state2))))
      (t (cond (fold-atom (setf (f-pos-atoms fold-atom) pos-lits)
                          (setf (f-neg-atoms fold-atom) neg-lits)))
         (return t)))
state3                   
(cond (interpretation (setq atom (cdar interpretation))
       (cond 
        ((<= (a-last-inference atom) inference)
         (cond 
          ((and $deletion (null deletion) 
                (null (a-falsum atom)) (eq (a-evaluated atom) 'l))
           (setq deletion nil)))
         (setq interpretation (cdr interpretation)) (go state3))
        (t (cond ((= (caar interpretation) 1)
                  (setq neg-lits (cons atom neg-lits)))
                 (t (setq pos-lits (cons atom pos-lits))))
         (cond 
          ((and $deletion (null deletion) 
                (null (a-falsum atom)) (eq (a-evaluated atom) 'l))
           (setq deletion atom)))
         (setq interpretation (cdr interpretation)) (go state3))))
      (t (cond (fold-atom (setf (f-pos-atoms fold-atom) pos-lits)
                          (setf (f-neg-atoms fold-atom) neg-lits)))
        (cond
         ((and (= $lemmata 2) $lemma-tag)
          (cond
           ((= fold-sign 1) 
            (setq pos-lits (cons (f-atom fold-atom) pos-lits)))
           (t (setq neg-lits (cons (f-atom fold-atom) neg-lits))))
          (setq clause (vector nil length pos-lits neg-lits length 0 0))
          (cond ((and $deletion deletion)
                 (setf (a-delete-lemmata deletion) 
                       (cons clause (a-delete-lemmata deletion)))))
          (setf (a-lemmata insertion) (cons clause (a-lemmata insertion)))))))
))

(defun attach-clauses (clauses)
(declare (special $lemmalist $max-lemma-store $lemma-store) 
         (fixnum $max-lemma-store $lemma-store))
(let (clause secondlast)
 (cond 
  ((and clauses (> $max-lemma-store 0)) (setq clause (car clauses))
   (setq $lemmalist (cons clause $lemmalist))
   (attach-clause-pos (c-pos-literals clause) clause)
   (attach-clause-neg (c-neg-literals clause) clause)
   (cond ((= $max-lemma-store $lemma-store) 
          (setq secondlast (secondlast $lemmalist))
          (detach-lemma-pos (c-pos-literals (cadr secondlast)))
          (detach-lemma-neg (c-neg-literals (cadr secondlast)))
          (rplacd secondlast nil))
          (t (setq $lemma-store (1+ $lemma-store))))
   (attach-clauses (cdr clauses)))
)))

(defun detach-lemma-pos (literals)
 (cond (literals (rem-last (a-last-pos-clauses (car literals)))
                 (detach-lemma-pos (cdr literals)))
))

(defun detach-lemma-neg (literals)
 (cond (literals (rem-last (a-last-neg-clauses (car literals)))
                 (detach-lemma-neg (cdr literals)))
))

(defun attach-clause-pos (literals clause)
(let (pos-last)
 (cond (literals (setq pos-last (a-last-pos-clauses (car literals)))
        (rplacd pos-last (cons clause (cdr pos-last)))
        (attach-clause-pos (cdr literals) clause))
)))

(defun attach-clause-neg (literals clause)
(let (neg-last)
 (cond (literals (setq neg-last (a-last-neg-clauses (car literals)))
        (rplacd neg-last (cons clause (cdr neg-last)))
        (attach-clause-neg (cdr literals) clause))
)))

(defun delete-clauses (clauses)
(declare (special $lemmalist))
(let (clause)
 (cond (clauses (setq clause (car clauses))
        (delete-clause-pos (c-pos-literals clause) clause)
        (delete-clause-neg (c-neg-literals clause) clause)
        (cond ((eq clause (car $lemmalist)) 
               (setq $lemmalist (cdr $lemmalist)))
;(;           (t (once-delete-predcdr $lemmalist (cdr $lemmalist) clause)))
              (t (once_delete_cdr $lemmalist clause)))
        (delete-clauses (cdr clauses)))
)))
 
(defun delete-clause-pos (literals clause)
 (cond
  (literals
   (cond
    ((and (svref (car literals) 1)
          (eq (car (svref (car literals) 1)) clause))
     (setf (svref (car literals) 1) (cdr (svref (car literals) 1))))
    ((svref (car literals) 1)
     (once_delete_cdr (svref (car literals) 1) clause)))
   (delete-clause-pos (cdr literals) clause))
))
 
(defun delete-clause-neg (literals clause)
 (cond
  (literals
   (cond
    ((and (svref (car literals) 2)
          (eq (car (svref (car literals) 2)) clause))
     (setf (svref (car literals) 2) (cdr (svref (car literals) 2))))
    ((svref (car literals) 2)
     (once_delete_cdr (svref (car literals) 2) clause)))
   (delete-clause-neg (cdr literals) clause))
))

(defun replace-clauses (clauses)
(declare (special $lemmalist $lemma-store $max-lemma-store)
         (fixnum $lemma-store $max-lemma-store))
(let (clause init-length1 lemma init-length2)
(declare (fixnum init-length1 init-length2))
 (cond
  (clauses (setq clause (car clauses))
           (setq init-length1 (c-init-length clause))
   (cond
    ($lemmalist (setq lemma (car $lemmalist))
                (setq init-length2 (c-init-length lemma))
     (cond ((or (< init-length1 init-length2)
                (and (= init-length1 init-length2)
                     (< (c-length clause) (c-length lemma))))
            (insert-clause-pos (c-pos-literals clause) clause)
            (insert-clause-neg (c-neg-literals clause) clause)
            (clause-insert-inv-predcdr $lemmalist (cdr $lemmalist)
                                       clause init-length1)
            (cond ((= $lemma-store $max-lemma-store)
                   (delete-clause-pos (c-pos-literals lemma) lemma)
                   (delete-clause-neg (c-neg-literals lemma) lemma)
                   (setq $lemmalist (cdr $lemmalist)))
                  (t (setq $lemma-store (1+ $lemma-store)))))
           ((< $lemma-store $max-lemma-store)
            (insert-clause-pos (c-pos-literals clause) clause)
            (insert-clause-neg (c-neg-literals clause) clause)
            (setq $lemma-store (1+ $lemma-store))
            (setq $lemmalist (cons clause $lemmalist)))))
    (t (insert-clause-pos (c-pos-literals clause) clause)
       (insert-clause-neg (c-neg-literals clause) clause)
       (setq $lemma-store (1+ $lemma-store))
       (setq $lemmalist (list clause))))
   (replace-clauses (cdr clauses)))
)))

(defun insert-clause-pos (literals clause)
(let (a-last-pos-clauses)
 (cond (literals 
        (setq a-last-pos-clauses (a-last-pos-clauses (car literals)))
        (clause-insert-predcdr a-last-pos-clauses (cdr a-last-pos-clauses) 
                               clause (c-init-length clause))
        (insert-clause-pos (cdr literals) clause))
)))

(defun insert-clause-neg (literals clause)
(let (a-last-neg-clauses)
 (cond (literals 
        (setq a-last-neg-clauses (a-last-neg-clauses (car literals)))
        (clause-insert-predcdr a-last-neg-clauses (cdr a-last-neg-clauses) 
                               clause (c-init-length clause))
        (insert-clause-neg (cdr literals) clause))
)))

(defmacro clause-insert-predcar (predcar clause init-length)
(declare (fixnum init-length))
 `(let (clauses f-clause) (setq clauses (car ,predcar))
   (cond 
    (clauses (setq f-clause (car clauses))
     (cond ((or (< ,init-length (c-init-length f-clause))
                (and (= ,init-length (c-init-length f-clause))
                     (<= (c-length ,clause) (c-length f-clause))))
            (rplaca ,predcar (cons ,clause clauses)))
           (t 
            (clause-insert-predcdr clauses (cdr clauses) 
                                           ,clause ,init-length))))
    (t (rplaca ,predcar (list ,clause)))
)))
   
(defun clause-insert-predcdr (predcdr clauses clause init-length)
 (declare (fixnum init-length))
 (cond 
  (clauses 
   (cond ((or (< init-length (c-init-length (car clauses)))
              (and (= init-length (c-init-length (car clauses)))
                   (<= (c-length clause) (c-length (car clauses)))))
          (rplacd predcdr (cons clause clauses)))
         (t (clause-insert-predcdr clauses (cdr clauses) 
                                   clause init-length))))
  (t (rplacd predcdr (list clause)))
))

(defmacro clause-insert-inv-predcar (predcar clause init-length)
(declare (fixnum init-length))
 `(let (clauses f-clause) (setq clauses (car ,predcar))
   (cond 
    (clauses (setq f-clause (car clauses))
     (cond ((or (> ,init-length (c-init-length f-clause))
                (and (= ,init-length (c-init-length f-clause))
                     (>= (c-length ,clause) (c-length f-clause))))
            (rplaca ,predcar (cons ,clause clauses)))
           (t 
            (clause-insert-inv-predcdr clauses (cdr clauses) 
                                               ,clause ,init-length))))
    (t (rplaca ,predcar (list ,clause)))
)))
   
(defun clause-insert-inv-predcdr (predcdr clauses clause init-length)
 (declare (fixnum init-length))
 (cond 
  (clauses 
   (cond ((or (> init-length (c-init-length (car clauses)))
              (and (= init-length (c-init-length (car clauses)))
                   (>= (c-length clause) (c-length (car clauses)))))
          (rplacd predcdr (cons clause clauses)))
         (t (clause-insert-inv-predcdr clauses (cdr clauses) 
                                       clause init-length))))
  (t (rplacd predcdr (list clause)))
))

(defmacro once-delete-vector (position atom list elem)
 `(cond (,list
         (cond ((eq (car ,list) ,elem) 
                (setf (svref ,atom ,position) (cdr ,list)))
               (t (once_delete ,list ,elem))))))
;((((;         (t (once-delete-predcdr ,list (cdr ,list) ,elem))))))
 
;(defun once-delete-predcdr (predcdr list elem)
; (cond (list
;        (cond ((eq (car list) elem) (setf (cdr predcdr) (cdr list)))
;              (t (once-delete-predcdr list (cdr list) elem))))
;))

(defun once_delete_list (list1 list2)
 (cond
  (list1
   (once_delete_list (rest list1) (once_delete (first list1) list2)))
  (t list2)))
 
(defmacro once_delete (elem list)
`(cond ((null ,list) nil)
       ((eq ,elem (car ,list)) (cdr ,list))
       (t (once_delete_cdr ,list ,elem) ,list)))
 
(defun once_delete_cdr (list elem)
 (cond ((cdr list)
        (cond ((eq elem (first (cdr list))) (setf (cdr list) (cddr list)))
              (t (once_delete_cdr (cdr list) elem))))))
 
(defun once_delete_pred (predlist list elem)
 (cond (list
        (cond ((eq (car list) elem) (setf (cdr predlist) (cdr list)))
              (t (once_delete_pred list (cdr list) elem))))))
 
;--------------model output------------------------------------------

(defmacro model (interpretation inference)
 (declare (special $models $verbose $output) (fixnum $verbose))
 `(cond ((= 1 $verbose) (terpri)
         (princ "Model found after ") (prin1 ,inference) 
        ; (setq $models (cons ,interpretation $models))
         (princ " inferences  ")  (terpri) nil)
        ((= 2 $verbose) (terpri)
         (princ "Model found after ") (prin1 ,inference) 
         (princ " inferences:  ") (terpri)
         (cond ((boundp '$output)
                (print-model $output ,interpretation (1- $output)))
               (t (print-model 10 ,interpretation 9))) (terpri))
         ;(t (princ "Model found after ") (prin1 ,inference) 
         ;   (princ " inferences  ") (terpri) nil)
))

(defun print-model (nr interpretation line)
 (declare (fixnum nr)) 
 (cond (interpretation
        (cond ((= 0 nr) (terpri) (setq nr line))
              (t (setq nr (1- nr))))
        (cond ((= 1 (caar interpretation)) (princ " "))
              (t (princ "~")))
        (prin1 (a-name (cdr (car interpretation)))) (princ " ");;***
        (print-model nr (cdr interpretation) line))
))

(defun init-inference (atoms atom)
 (cond (atoms (setq atom (car atoms)) (setf (svref atom 3) 0)
              (init-inference (cdr atoms) nil))))

(defun init-inf () 
 (declare (special $atomlist $models))
 (init-inference $atomlist nil) (setq $models nil))

(defun secondlast (list)
 (cond ((cddr list) (secondlast (cdr list)))
       (t list)
))

(defun rem-last (list) ; list must be non-empty
 (cond ((cddr list) (rem-last (cdr list)))
       (t (rplacd list nil))
))

(defmacro d-max-atom (atoms)
 `(prog (m-pos m-neg atom)
   (declare (fixnum m-pos m-neg))
   (setq atom (car ,atoms))
   (setq m-pos (a-pos-frequency atom)) 
   (setq m-neg (a-neg-frequency atom))
   (cond ((< m-pos m-neg) 
          (return (d-max-atom-atoms (cdr ,atoms) atom m-neg m-pos)))
         (t (return (d-max-atom-atoms (cdr ,atoms) atom m-pos m-neg))))))

(defun d-max-atom-atoms (atoms atom max min)
 (declare (fixnum max min))
 (prog (pos neg new-atom)
  (declare (fixnum pos neg))
  loop
  (cond
   (atoms (setq new-atom (car atoms))
    (cond 
     ((a-evaluated new-atom) 
      (return (d-max-atom-atoms (cdr atoms) atom max min)))
     (t (setq pos (a-pos-frequency new-atom))
        (setq neg (a-neg-frequency new-atom))
      (cond
       ((>= pos neg)
        (cond
         ((> max pos) (setq atoms (cdr atoms)) (go loop)) 
         ((= max pos)
          (cond ((>= min neg) (setq atoms (cdr atoms)) (go loop))
                (t (setq atoms (cdr atoms)) (setq max pos) (setq min neg)
                   (setq atom new-atom) (go loop))))
         (t (setq atoms (cdr atoms)) (setq max pos) (setq min neg)
            (setq atom new-atom) (go loop))))
       ;; in the following: neg > pos:
       ((> max neg) (setq atoms (cdr atoms)) (go loop))
       ((= max neg)
        (cond ((>= min pos) (setq atoms (cdr atoms)) (go loop))
              (t (setq atoms (cdr atoms)) (setq max neg) (setq min pos)
                 (setq atom new-atom) (go loop))))
       (t (setq atoms (cdr atoms)) (setq max neg) (setq min pos)
          (setq atom new-atom) (go loop))))))
   (t (return atom)))))

(defun random-atom (atom-list)
(declare (special $atom-nr) (fixnum $atom-nr))
 (prog (nr atomlist)
(declare (fixnum nr))
  (setq nr (random $atom-nr))
  loop
  (cond ((setq atomlist (nthcdr nr atom-list))
         (cond ((setq atomlist (rem-prefix atomlist))
                (return (car atomlist)))
               (t (return (car atom-list)))))
        ((evenp nr) (setq nr (/ nr 2)) (go loop))
        (t (setq nr (/ (1- nr) 2)) (go loop))
)))


; ---------------------------------------------------------------------
; outputs the clauses used in the proof

(defun output-clauses (clauselist)
 (cond (clauselist 
        (cond ((> (c-usages (car clauselist)) 0) 
               (terpri)
               (print-clause-pos (c-pos-literals (car clauselist)))
               (print-clause-neg (c-neg-literals (car clauselist)))
               (princ " /* ") (prin1 (c-usages (car clauselist)))
               (princ " usages of clause ") 
               (prin1 (c-number (car clauselist))) (princ " */")
        ))
        (output-clauses (cdr clauselist)))
))

(defun print-clause-pos (pos)
 (cond (pos (prin1 (a-name (car pos)))
        (cond ((cdr pos) (princ "; ")
               (print-clause-pos (cdr pos)))
              (t (princ " <- ")))) 
       (t (princ "<- "))
))

(defun print-clause-neg (pos)
 (cond (pos (prin1 (a-name (car pos)))
        (cond ((cdr pos) (princ ", ")
               (print-clause-neg (cdr pos)))
              (t (princ "."))))
       (t (princ "."))
))


;;;*************************************************************************
;;;module name: cl- compiler.l
;;;
;;;description: generates data structures for cl-semprop.l
;;;
;;;remarks    : the input data structure $formula is destructively modified 
;;;		and the previously unbound predicate symbols in the input 
;;;		formula are setqed in the compilation process. 
;;;		therefore, if another formula is to be processed or 
;;;		recompilation is wanted, (clear) has to be called and the 
;;;		input file has to be loaded (again).
;;;
;; -------------------------------------------------------------------------
;; input data: 
;; -------------------------------------------------------------------------
;; 
;; global variables:
;; $formula: list of input-clauses:
;; 
;; data structures:
;; input-clause: list of two clause-halfs (positive and negative literals)
;;               plus possibly the clause number
;; clause-half: list of symbols
;; -------------------------------------------------------------------------

;; -------------------------------------------------------------------------
;; output data:
;; -------------------------------------------------------------------------
;; -------------------------------------------------------------------------
;; the following three new global variables are created:
;; $clauselist: list of clauses to be refuted
;; $atomlist: list of atoms in formula
;; $units: list of unit clauses
;;
;; -------------------------------------------------------------------------
;; main procedure: processes list of input-clauses and generates the 
;; described output data structures
;; -------------------------------------------------------------------------

(defun comp () 
 (cond ((and (boundp '$compiled) $compiled)
        (princ 
         "input formula destroyed. call `(clear)' and load input formula!")
         nil)
       ($formula (comp-semp) (eval-comp) (make-last $atomlist)
                 (setq $compiled t))
       (t (princ "empty input formula. load input formula!") nil)
))
;; -------------------------------------------------------------------------
(defun comp-semp () 
 (setq $units nil) (setq $atomlist nil) (setq $clauselist nil)
 (compile-to-semprop $formula 1)
 (setq $atomlist (nreverse $atomlist)) (setq $units (nreverse $units))
)

(defun compile-to-semprop (input-clauses clause-nr) 
 (cond (input-clauses 
        (compile-input-clause (car input-clauses) clause-nr)
        (compile-to-semprop (cdr input-clauses) (1+ clause-nr)))))

(defun list2set (list)
 (cond ((cdr list)
        (cond ((element (car list) (cdr list)) (list2set (cdr list)))
              (t (cons (car list) (list2set (cdr list))))))
       (t list)))

(defun element (elem list)
 (cond ((and list (eq elem (car list))) list)
       (list (element elem (cdr list)))))

(defun no_intersection (list1 list2)
 (cond (list2
        (cond ((element (first list2) list1) (first list2))
              (t (no_intersection list1 (rest list2)))))
       (t t)))

(defun compile-input-clause (input-clause clause-nr) 
 (prog (cl-length clause cl-nr poslits neglits)
  (setq cl-nr (nth 3 input-clause))
  (setq poslits (list2set (nth 0 input-clause)))
  (setq neglits (list2set (nth 1 input-clause)))
  (cond ((no_intersection poslits neglits) ; no tautology
         (cond (cl-nr nil) (t (setq cl-nr clause-nr)))
         (setq cl-length (+ (length poslits) (length neglits)))
         (setq clause 
	       (vector nil cl-length poslits neglits cl-length 0 cl-nr))
         (compile-input-clause-half poslits nil cl-length clause 1) 
         (compile-input-clause-half neglits nil cl-length clause -1) 
         (setq $clauselist (cons clause $clauselist))
         (cond ((= 1 cl-length) (setq $units (cons clause $units)))))
)))

(defun compile-input-clause-half 
 (clause-half atom-name length clause sign)
 (declare (fixnum sign))
 (cond (clause-half (setq atom-name (car clause-half))
         (cond 
           ((boundp atom-name) (setq atom-name (symbol-value atom-name))
            (cond ((= sign 1)
                   (setf (svref atom-name 4)
                         (1+ (a-pos-frequency atom-name)))
                   (cond 
                     ((a-pos-clauses atom-name) 
                      (cond 
                       ((boundp 'inverse)
                        (insert-clause-inverse 
                         (a-pos-clauses atom-name) length clause))
                       (t (insert-clause-inverse 
                           (a-pos-clauses atom-name) length clause))))
                     (t (setf (a-pos-clauses atom-name) (list clause)))))
                  ((a-neg-clauses atom-name) 
                   (cond 
                    ((boundp 'inverse)
                     (insert-clause-inverse 
                      (a-neg-clauses atom-name) length clause))
                    (t (insert-clause
                        (a-neg-clauses atom-name) length clause)))
                   (setf (a-neg-frequency atom-name)
                         (1+ (a-neg-frequency atom-name))))
                  (t (setf (a-neg-clauses atom-name) (list clause))
                     (setf (a-neg-frequency atom-name)
                           (1+ (a-neg-frequency atom-name)))))
            (setq atom-name (a-name atom-name)))
           ((= sign 1) 
            (set atom-name 
              (vector nil (list clause) nil 0 1 0 nil nil nil atom-name 
                      nil nil nil nil 0 0))
            (setq $atomlist (cons (symbol-value atom-name) $atomlist)))
           (t 
            (set atom-name 
               (vector nil nil (list clause) 0 0 1 nil nil nil atom-name 
                       nil nil nil nil 0 0))
            (setq $atomlist (cons (symbol-value atom-name) $atomlist))))
            (compile-input-clause-half 
                 (cdr clause-half) nil length clause sign))))
          
(defun insert-clause (clauses length clause)
 (declare (fixnum length))
 (cond ((< length (c-length (car clauses)))
        (splice-in clause clauses))
       ((cdr clauses)
        (insert-clause (cdr clauses) length clause))
       (t (rplacd clauses (list clause)))))

(defun insert-clause-inverse (clauses length clause)
 (declare (fixnum length))
 (cond ((>= length (c-length (car clauses)))
        (splice-in clause clauses))
       ((cdr clauses)
        (insert-clause-inverse (cdr clauses) length clause))
       (t (rplacd clauses (list clause)))))

(defun splice-in (element non-empty-list)
 (rplacd non-empty-list (cons (car non-empty-list) (cdr non-empty-list)))
 (rplaca non-empty-list element))

;;;*************************************************************************
;;;module name: eval-compiler.l
;;;
;;;description: generates final data structures for cl-semprop.l by modifying 
;;;		data structures generated by comp-semp as follows:
;;;             all symbols in c-pos-literals and c-neg-literals in every
;;;		clause are replaced by their values, respectively
;;;*************************************************************************

;; -------------------------------------------------------------------------
;; input data: 
;; -------------------------------------------------------------------------
;; 
;; global variable:
;; $atomlist: list of atoms
;; $clauselist: list of clauses
;; 
;; -------------------------------------------------------------------------
;; output data:
;; -------------------------------------------------------------------------
;; accordingly modified global variables
;; $atomlist: list of atoms
;; $clauselist: list of clauses
;; 
;; -------------------------------------------------------------------------
;; relevant data structures:
;; clause: vector of seven elements
;; 2nd element: current clause length (integer)
;; 3rd element: list of positive atoms
;;
;; atom: vector of fourteen elements
;; 2nd element: clauses with positive occurrence of atom
;; 3rd element: clauses with negative occurrence of atom

(defun eval-comp ()
(eval-comp-clauselist $clauselist))

(defun eval-comp-atomlist (atoms) 
 (cond (atoms (replace-by-values (a-pos-clauses (car atoms)))
              (replace-by-values (a-neg-clauses (car atoms)))
              (eval-comp-atomlist (cdr atoms)))))

(defun eval-comp-clauselist (clauses) 
 (cond (clauses (replace-by-values (c-pos-literals (car clauses)))
                (replace-by-values (c-neg-literals (car clauses)))
                (eval-comp-clauselist (cdr clauses))))) 

(defun replace-by-values (symb-list)
 (cond (symb-list (rplaca symb-list (symbol-value (car symb-list)))
                  (replace-by-values (cdr symb-list)))))

(defun make-last (atomlist)
(prog (atom)
 (cond (atomlist (setq atom (car atomlist))
        (setf (svref atom 10) (last (a-pos-clauses atom)))
        (setf (svref atom 11) (last (a-neg-clauses atom)))
        (make-last (cdr atomlist)))
)))

(defun clear () 
 (cond ((boundp '$atomlist) (rmakunbound-name $atomlist)))
 (setq $lemma-store 0) (setq $lemmalist nil)
 (setq $formula nil) (setq $compiled nil))

(defun rmakunbound-name (atomlist)
 (cond (atomlist (makunbound (a-name (car atomlist)))
                             (rmakunbound-name (cdr atomlist)))))

(defun lprove (filename) (clear) (load filename) (comp) (prove) )
(defun ltprove (filename) (clear) (load filename) (comp) (time (prove)))
