%###########################################################################
%# cmd_line.pl
%# - retrieval and processing of command line arguments
%# - command line is assumed to be arranged as follows:
%#   <progname> { -<switch> {<switch-argument>} } <filename>
%#   where <filename> can also be '-' to indicate standard input
%###########################################################################
%# predicates to be used from outside:
%#	- process_cmd_line/1
%# dependent on
%#	- config.pl
%###########################################################################
%#        Max Moser                         Automated Reasoning Research Group
%# Email: moser@informatik.tu-muenchen.de     Lehrstuhl VIII f"ur Informatik
%# Tel.:  +49-89/521096			     Technische Universit"at M"unchen
%# Fax.:  +49-89/526502			         D-80290 M"unchen, Germany
%# http://wwwjessen.informatik.tu-muenchen.de/~moser
%###########################################################################

:- dynamic cmd_arg/2.
:- dynamic infile/1.
:- dynamic modswitch/1.


%*****************************************************************************
%* process_cmd_line( -SwitchList )
%*****************************************************************************
%* - retrieve, parse and arrange command line arguments
%* - show usage and exit when error found
%* - assert command line arguments as `cmd_arg(...)'
%* - assert file name as `infile(...)'
%* - write comment on stdout
%*****************************************************************************

process_cmd_line( SwitchList) :-
		% first retrieve file name and all the switches
	get_file_name( FileName ),
	get_switch_list( SwitchList ),
		% then check them and assert facts about them
	remember_file_name( FileName ),
	remember_switches( SwitchList ),
	write("\n").

process_cmd_line( '' ) :-
		% in case of error, show usage and stop
	usage,
	exit(1).

%*****************************************************************************


%*****************************************************************************
%* get_file_name( -FileName )
%*****************************************************************************
%* - find out whether last argument is file name or '-'
%* - fail if neither is the case
%* - convert string to atom
%*****************************************************************************

get_file_name( FileName ) :-
		% test whether there are enough arguments
	argc( CmdLineLength ),
	CmdLineLength > 1,
		% even if there are more arguments, make sure it is not a
		% dummy generated by HPUX
	argv( 1, N ),
	N \== "",
		% retrieve last argument
	LastCmdLinePos is CmdLineLength - 1,
	argv( LastCmdLinePos, FileNameString ),
		% test whether last argument is no switch
	name( FileNameString, FileNameAscii ),
	name( '-', [MinusAscii] ),
	FileNameAscii \= [MinusAscii,_|_],
	name( FileName, FileNameAscii ).

%*****************************************************************************


%*****************************************************************************
%* get_switch_list( -SwitchList )
%*****************************************************************************
%* - retrieve and arrange list of command line switches
%*   SwitchList has the form [...,[-switch],...,[-switch,arg,arg],...]
%*****************************************************************************

get_switch_list( SwitchList ) :-
		% find position of last switch
	argc( CmdLineLength ),
	SwitchPos is CmdLineLength - 2,
		% retrieve unstructured list of switches and arguments
	retrieve_cmdline( 1, SwitchPos, PlainSwitchList ),
		% arrange PlainSwitchList to structured SwitchList
	arrange_switch_list( PlainSwitchList, SwitchList ).

%*****************************************************************************


%*****************************************************************************
%* retrieve_cmdline( +FromPos, +ToPos, -CmdLineArgList )
%*****************************************************************************
%* - collect all the command line arguments from FromPos to ToPos
%*****************************************************************************

retrieve_cmdline( FromPos, ToPos, [] ) :-
		% no result if the range is empty
	FromPos > ToPos,
	!.

retrieve_cmdline( FromPos, ToPos, [CmdLineArg|CmdLineArgListRest] ) :-
		% if range is not empty retrieve command line argument
	argv( FromPos, CmdLineArg ),
		% continue with remaining command line arguments
	NewFromPos is FromPos + 1,
	retrieve_cmdline( NewFromPos, ToPos, CmdLineArgListRest ).

%*****************************************************************************


%*****************************************************************************
%* arrange_switch_list( +PlainSwitchList, -SwitchList )
%*****************************************************************************
%* - arrange PlainSwitchList so that SwitchList has the form 
%*   [...,[-switch],...,[-switch,arg,arg],...]
%*****************************************************************************

arrange_switch_list( [], [] ).

arrange_switch_list( [Switch|PlainSwitchList], [SwitchStruct|SwitchList] ) :-
		% test if command line argument is switch
	name( '-', [MinusAscii] ),
	name( Switch, [MinusAscii|_] ),
		% collect all arguments belonging to current switch
	arrange_switch_list( [Switch], PlainSwitchList, 
				SwitchStruct, PlainSwitchListRest ),
		% continue with remaining plain switch list
	arrange_switch_list( PlainSwitchListRest, SwitchList ).


arrange_switch_list( SwitchStruct, [], SwitchStruct, [] ).
		% if no more switches then we are done

arrange_switch_list( SwitchStruct, [Sw|PlainSwitchList], 
			SwitchStructResult, PlainSwitchListRest ) :-
		% if Sw is a number (eventually starting with a '-') ...
	name( Sw, SwAscii ),
	name( SwInt, SwAscii ),
	number( SwInt ),
	!,
		% ... then extend SwitchStruct and continue parsing
	append( SwitchStruct, [SwInt], NewSwitchStruct ),
	arrange_switch_list( NewSwitchStruct, PlainSwitchList,
			SwitchStructResult, PlainSwitchListRest ).

arrange_switch_list( SwitchStruct, [Sw|PlainSwitchList], 
			SwitchStruct, [Sw|PlainSwitchList] ) :-
		% if Sw starts with a '-' (but is no number) ...
	name( '-', [MinusAscii] ),
	name( Sw, [MinusAscii|_] ),
		% ... then stop arranging switch struct
	!.

arrange_switch_list( SwitchStruct, [Sw|PlainSwitchList], 
			SwitchStructResult, PlainSwitchListRest ) :-
		% else, extend SwitchStruct and continue parsing
	name( Sw, SwAscii ),
	name( SwAtom, SwAscii ),
	append( SwitchStruct, [SwAtom], NewSwitchStruct ),
	arrange_switch_list( NewSwitchStruct, PlainSwitchList,
			SwitchStructResult, PlainSwitchListRest ).

%*****************************************************************************


%*****************************************************************************
%* remember_file_name( +InFile )
%*****************************************************************************
%* - save input file name as asserted fact infile(InFile)
%* - write infile as comment to stdout
%*****************************************************************************

remember_file_name( InFile ) :-
	assert( infile(InFile) ),
	write('# *** infile '), writeln(InFile).

%*****************************************************************************


%*****************************************************************************
%* remember_switches( +ArrangedSwitchList )
%*****************************************************************************
%* - process arranged switches, i.e. save switches and argument info
%* - invoke predicates of filters
%* - process_switch/1 is defined and configured in config.pl
%*****************************************************************************

remember_switches( [] ).

remember_switches( [ArrangedSwitch|RestArrangedSwitches] ) :-
	process_system_switch( ArrangedSwitch ),
	!,
	remember_switches( RestArrangedSwitches ).

remember_switches( [ArrangedSwitch|RestArrangedSwitches] ) :-
        process_modification_switch( ArrangedSwitch ),
	process_switch( ArrangedSwitch ),
	!,
	remember_switches( RestArrangedSwitches ).

remember_switches( [ArrangedSwitch|RestArrangedSwitches] ) :-
	process_switch( ArrangedSwitch ),
	!,
	remember_switches( RestArrangedSwitches ).

remember_switches( [ArrangedSwitch|_] ) :-
	write( stderr, '*** unknown switch/switch+argument combination: ' ), 
	writeln( stderr, ArrangedSwitch ),
	usage,
	exit(1).

%*****************************************************************************


%*****************************************************************************
%* process_system_switch( +ArrangedSwitch )
%*****************************************************************************
%* - process `built in' switches
%*****************************************************************************

process_system_switch( ["-tptpin"] ) :-
	assert( cmd_arg(tptpin,[]) ).

process_system_switch( ["-tptpout"] ) :-
	assert( cmd_arg(tptpout,[]) ).

process_system_switch( ["-verbose"] ) :-
	assert( cmd_arg(verbose,[]) ).

%*****************************************************************************


%*****************************************************************************
%* process_modification_switch( +ArrangedSwitch )
%*****************************************************************************
%* - sets modification switches
%*****************************************************************************

process_modification_switch( ["-stemod"] ) :-
	assert( modswitch(stemod) ).

process_modification_switch( ["-moserstemod"] ) :-
	assert( modswitch(moserstemod) ).

process_modification_switch( ["-brandstemod"] ) :-
	assert( modswitch(brandstemod) ).

process_modification_switch( ["-ceemod"] ) :-
	assert( modswitch(ceemod) ).

%*****************************************************************************
