%############################################################################
%# mod_stemod.pl                                                            #
%# - STE modifications from                                                 #
%#                         [Brand;75]                                       #
%#                         [MoserSteinbach;97] (original & revised version) #
%#                         [BachmairGanzingerVoronkov:MPI;97]               #
%############################################################################
%# predicates to be used from outside:                                      #
%#	- addrefleq/4                                                       #
%#	- smod/4                                                            #
%#      - snmod/4                                                           #
%#      - tmod/4                                                            #
%#      - b_tmod/4                                                          #
%#	- tpmod/4                                                           #
%#      - tpcmod/4                                                          #
%#	- tnmod/4                                                           #
%#      - tncmod/4                                                          #
%#	- remod/4                                                           #
%#      - emod/4                                                            #
%#      - b_emod/4                                                          #
%#	- mlmod/4                                                           #
%#      - mmod/4                                                            #
%#      - process_switch_stemod/1                                           #
%#	- process_switch_moserstemod/1                                      #
%#	- process_switch_brandstemod/1                                      #
%#	- process_switch_ceemod/1                                           #
%#      - test_switches_stemod/0                                            #
%#	- test_switches_moserstemod/0                                       #
%#	- test_switches_brandstemod/0                                       #
%#	- test_switches_ceemod/0                                            #
%# dependent on                                                             #
%#	- mylib.pl                                                          #
%#	- clause info facts generated by siginfo.pl                         #
%############################################################################
%#        Joachim Steinbach              Automated Reasoning Research Group #
%# Email: steinbac@informatik.tu-muenchen.de Lehrstuhl VIII f"ur Informatik #
%# Tel.:  +49-89/289-27913		   Technische Universit"at M"unchen #
%# Fax.:  +49-89/289-27902		          D-80290 M"unchen, Germany #
%# http://wwwjessen.informatik.tu-muenchen.de/~steinbac                     #
%############################################################################


%****************************************************************************
%* addrefleq( problemwise, +ClauseList, +SwitchList, -NewClauseList )       *
%****************************************************************************
%* - just add equal(X,X) as first clause to clause list                     *
%****************************************************************************

addrefleq( problemwise, ClauseList, _, [ReflEq|ClauseList] ) :-
		% just construct reflexivity axiom
	ReflEq = rule( reflexivity, axiom, [++equal(X,X)], [], ['X'=X] ).


%****************************************************************************
%* mmod( clausewise, +Clause, +SwitchList, -NewClauseList )                 *
%****************************************************************************
%* - general elimination of monotonicity:                                   *
%*   C(s)  ==>  s=/=x v C(x)                                                *
%*   where s contains more than one operator and x is new wrt. C            *
%****************************************************************************

mmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

mmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

mmod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	m_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

mmod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* remod( clausewise, +Clause, +SwitchList, -NewClauseList )                *
%****************************************************************************
%* - partial elimination of reflexivity, i.e. of x =/= y:                   *
%*   x=/=y v C  ==>  C  with x=y                                            *
%****************************************************************************

remod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

remod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

remod( clausewise, Clause, _, NewClauseList ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	re_modif( LitList, NewLitList ),
	addinfo2newclauselist( Name, Type, [NewLitList], Cons, Dict, 
							     NewClauseList ).

remod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* smod( clausewise, +Clause, +SwitchList, -NewClauseList )                 *
%****************************************************************************
%* - elimination of symmetry:                                               *
%*   s=t v C  ==>  s=t v C                                                  *
%*   s=t v C  ==>  t=s v C                                                  *
%****************************************************************************

smod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

smod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

smod( clausewise, Clause, _, NewClauseList ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	s_modif( LitList, NewLitListList ),
	addinfo2newclauselist( Name, Type, NewLitListList, Cons, Dict, 
		       				             NewClauseList ).

smod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* snmod( clausewise, +Clause, +SwitchList, -NewClauseList )                *
%****************************************************************************
%* - perform Sn-modification:                                               *
%*   s=/=t v C  ==>  s=/=t v C  if s is not a variable                      *
%*   s=/=t v C  ==>  t=/=s v C  if s is a variable and t not                *
%****************************************************************************

snmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

snmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

snmod( clausewise, Clause, _, NewClauseList ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	sn_modif( LitList, NewLitList ),
	addinfo2newclauselist( Name, Type, [NewLitList], Cons, Dict, 
							     NewClauseList ).

snmod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* tmod( clausewise, +Clause, +SwitchList, -NewClauseList )                 *
%****************************************************************************
%* - perform T-modification:                                                *
%*   s=t v C  ==>  s=z v t=/=z v C  if t is not a variable and              *
%*                                     z is new in s=t v C                  *
%****************************************************************************

tmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

tmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

tmod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	t_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

tmod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* b_tmod( clausewise, +Clause, +SwitchList, -NewClauseList )               *
%****************************************************************************
%* - perform Brand's T-modification which is the same as the above tmod     *
%*   except that equations with variables on the right-hand sides must also *
%*   be split up                                                            *
%****************************************************************************

b_tmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

b_tmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

b_tmod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	b_t_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

b_tmod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* tpmod( clausewise, +Clause, +SwitchList, -NewClauseList )                *
%****************************************************************************
%* - perform t-modification for positive equations (identical to tmod)      *
%****************************************************************************

tpmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

tpmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

tpmod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	tp_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

tpmod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* tpcmod( clausewise, +Clause, +SwitchList, -NewClauseList )               *
%****************************************************************************
%* - perform T-modification for positive equations adding constraints:      *
%*   (s=t v C) . A  ==>  (s=z v t=/=z v C) . (A ^ s>z ^ t>=z)               *
%****************************************************************************

tpcmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

tpcmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

tpcmod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
        setval( rescnt, 0 ),
	tpc_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

tpcmod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* tnmod( clausewise, +Clause, +SwitchList, -NewClauseList )                *
%****************************************************************************
%* - perform T-modification for negative equations:                         *
%*   s=/=t v C  ==>  s=/=z v t=/=z v C  if t is not a variable              *
%****************************************************************************

tnmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

tnmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

tnmod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	tn_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

tnmod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* tncmod( clausewise, +Clause, +SwitchList, -NewClauseList )               *
%****************************************************************************
%* - perform T-modification for negative equations adding constraints:      *
%*   (s=/=t v C) . A  ==>  (s=/=z v t=/=z v C) . (A ^ s>=z ^ t>=z)          *
%****************************************************************************

tncmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

tncmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

tncmod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
        setval( rescnt, 0 ),
	tnc_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

tncmod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* emod( clausewise, +Clause, +SwitchList, -NewClauseList )                 *
%****************************************************************************
%* - perform E-modification which corresponds to the application of remod,  *
%*   tnmod and mlmod                                                        *
%****************************************************************************

emod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

emod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

emod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	e_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

emod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* b_emod( clausewise, +Clause, +SwitchList, -NewClauseList )               *
%****************************************************************************
%* - perform Brand's E-modification which is the same as the above emod     *
%*   except that                                                            *
%*   o the root of right-hand sides in negative equations needs not to be   *
%*     replaced, and                                                        *
%*   o inferences on the left-hand side of x=/=y (where x, y are variables) *
%*     must be performed.                                                   *
%****************************************************************************

b_emod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

b_emod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

b_emod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	b_e_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

b_emod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* mlmod( clausewise, +Clause, +SwitchList, -NewClauseList )                *
%****************************************************************************
%* - elimination of monotonicity for left-hand sides which is the same as   *
%*   mmod except that only letf-hand sides of equations must be considered  *
%****************************************************************************

mlmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify equality clauses
	Clause = clause( _, _, LitList, _, _ ),
	unwanted_equality_axiom( LitList, [r,s,t,e] ),
	!.

mlmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

mlmod( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause( Name, Type, LitList, Cons, Dict ),
	!,
	ml_modif( LitList, NewTmpLitList, PreDict ),
	predict2dict( PreDict, Dict, NewVarDict ),
	append( Dict, NewVarDict, NewDict ),
	gen_companion_clause( NewTmpLitList, NewLitList ),
	NewClause = clause( Name, Type, NewLitList, Cons, NewDict ).

mlmod( clausewise, Clause, _, [Clause] ).


%****************************************************************************
%* m_modif( +Clause, -E_Clause, -PreDict )                                  *
%****************************************************************************
%* General Flattening                                                       *
%****************************************************************************

m_modif( LitList, ModLitList, PreDict ) :-
	m_modif_literal_list( LitList, ModLitList, PreDict ).


m_modif_literal_list( [], [], [] ).

m_modif_literal_list( [Lit|RestLiterals], ModLiterals, PreDict ) :-
	m_modif_literal( Lit, ModLitList, PreDictLit ),
	m_modif_literal_list( RestLiterals, ModRestLiterals, PreDictRest ),
	append( ModLitList, ModRestLiterals, ModLiterals ),
	append( PreDictLit, PreDictRest, PreDict ).


m_modif_literal( ++equal(L,R), [++equal(ModL,ModR)], PreDict ) :-
	!,
	m_modif_root_term( L, ModL, PreDictL ),
	m_modif_root_term( R, ModR, PreDictR ),
	append( PreDictL, PreDictR, PreDict ).

m_modif_literal( --equal(L,R), [--equal(ModL,ModR)], PreDict ) :-
	!,
	m_modif_root_term( L, ModL, PreDictL ),
	m_modif_root_term( R, ModR, PreDictR ),
	append( PreDictL, PreDictR, PreDict ).

m_modif_literal( ++Literal, [++ModLiteral], PreDict ) :-
	m_modif_root_term( Literal, ModLiteral, PreDict ).

m_modif_literal( --'$lpo'(L,R,Res), [--'$lpo'(L,R,Res)], [] ) :-
	!.

m_modif_literal( --'$lpoeq'(L,R,Res), [--'$lpoeq'(L,R,Res)], [] ) :-
	!.

m_modif_literal( --Literal, [--ModLiteral], PreDict ) :-
	m_modif_root_term( Literal, ModLiteral, PreDict ).


m_modif_root_term( Term, ModTerm, RealPreDict ) :-
	m_modif_term( Term, PreModTerm, PreDict ),
	( ext_var(PreModTerm) ->
		( ModTerm=PreModTerm, RealPreDict=[] );
		( PreModTerm=(_->ModTerm), PreDict=[_|RealPreDict] )
	).


m_modif_term( Var, Var, [] ) :-
	ext_var( Var ),
	!.

m_modif_term( Term, (NewLeft -> NewRight), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Funct,Subterm,S],
        cmd_arg( sortsnd, Sorts ),
        member( Funct, Sorts ),
        Subterm =.. [Functor|Args],
	m_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ),
        NewLeft =.. [Funct,Ref,S],
        NewRight =.. [Funct,EModTerm,S],
        !.

m_modif_term( Term, (NewLeft -> NewRight), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Funct,S,Subterm],
        cmd_arg( sortfst, Sorts ),
        member( Funct, Sorts ),
        Subterm =.. [Functor|Args],
	m_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ),
        NewLeft =.. [Funct,S,Ref],
        NewRight =.. [Funct,S,EModTerm],
        !.

m_modif_term( Term, (Ref -> EModTerm), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Functor|Args],
	m_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ).


m_modif_term_list( [], [], [] ).

m_modif_term_list( [Var|RestArgs], [Var|EModRestArgs], PreDict ) :-
	ext_var( Var ),
	!,
	m_modif_term_list( RestArgs, EModRestArgs, PreDict ).

m_modif_term_list( [Arg|RestArgs], [EModArg|EModRestArgs], PreDict ) :-
	m_modif_term( Arg, EModArg, PreDict1 ),
	m_modif_term_list( RestArgs, EModRestArgs, PreDictRest ),
	append( PreDict1, PreDictRest, PreDict).


%****************************************************************************
%* re_modif( +Clause, -S_ClauseList )                                       *
%****************************************************************************
%* R-modification: remove reflexivity                                       *
%****************************************************************************

re_modif( [--equal(L,R)|RestClause], R_RestClause ) :-
	ext_var(L), ext_var(R),
	!,
	L=R,
        re_modif( RestClause, R_RestClause ).

re_modif( [Lit|Clause], [Lit|R_Clause] ) :-
        re_modif( Clause, R_Clause ).

re_modif( [], [] ).


%****************************************************************************
%* s_modif( +Clause, -S_ClauseList )                                        *
%****************************************************************************
%* S-modification: introduce symmetrical positive equations                 *
%****************************************************************************

s_modif( [++equal(L,R)|RestClause], S_ClauseList ) :-
	L == R,
	!,
	s_modif( RestClause, S_RestClauseList ),
	prependtoall( ++equal(L,R), S_RestClauseList, S_ClauseList ).

s_modif( [++equal(L,R)|RestClause], S_ClauseList ) :-
	!,
	s_modif( RestClause, S_RestClauseList ),
	prependtoall( ++equal(L,R), S_RestClauseList, PrependL ),
	prependtoall( ++equal(R,L), S_RestClauseList, PrependR ),
	append( PrependL, PrependR, S_ClauseList ).

s_modif( [Literal|RestClause], S_ClauseList ) :-
	s_modif( RestClause, S_RestClauseList ),
	prependtoall( Literal, S_RestClauseList, S_ClauseList ).

s_modif( [], [[]] ).
	

%****************************************************************************
%* sn_modif( +Clause, -S_ClauseList )                                       *
%****************************************************************************
%* Sn-modification: move variables in negative equations to the right       *
%****************************************************************************

sn_modif( [], [] ).

sn_modif( [--equal(L,R)|RestClause], [--equal(R,L)|Sn_RestClause] ) :-
	ext_var( L ),
        not(ext_var( R )),
        !,
	sn_modif( RestClause, Sn_RestClause ).

sn_modif( [Lit|RestClause], [Lit|Sn_RestClause] ) :-
	sn_modif( RestClause, Sn_RestClause ).


%****************************************************************************
%* t_modif( +Clause, -T_ClauseList, -T_PreDict )                            *
%****************************************************************************
%* - compile transitivity                                                   *
%****************************************************************************

t_modif( [], [], [] ).

t_modif( [++equal(L,R)|RestClause], [++equal(L,R)|T_RestClause], PreDict ) :-
	var( R ),
	!,
	t_modif( RestClause, T_RestClause, PreDict ).

t_modif( [++equal(L,R)|RestClause], [++equal(L,(Ref->R))|T_RestClause], 
						[(Funct,Ref)|PreDict]) :-
	!,
	R =.. [Funct|_],
	getnewvar( Ref ),
	t_modif( RestClause, T_RestClause, PreDict ).

t_modif( [--equal(L,R)|RestClause], [--equal(L,R)|T_RestClause], PreDict ) :-
	!,
	t_modif( RestClause, T_RestClause, PreDict ).

t_modif( [Literal|RestClause], [Literal|T_RestClause], PreDict ) :-
	t_modif( RestClause, T_RestClause, PreDict ).


%****************************************************************************
%* b_t_modif( +Clause, -T_ClauseList, -T_PreDict )                          *
%****************************************************************************
%* - compile transitivity for Brand's T-modification                        *
%****************************************************************************

b_t_modif( [], [], [] ).

b_t_modif( [++equal(L,R)|RestClause], [++equal(L,(Ref->R))|T_RestClause], [(R,Ref)|PreDict] ) :-
	var( R ),
	!,
        getnewvar( Ref ),
	b_t_modif( RestClause, T_RestClause, PreDict ).

b_t_modif( [++equal(L,R)|RestClause], [++equal(L,(NewSubterm->R))|T_RestClause], [(R,Ref)|PreDict] ) :-
	ext_var( R ),
	R =.. [Funct,_,S],
	!,
        NewSubterm =.. [Funct,Ref,S],
	getnewvar( Ref ),
	b_t_modif( RestClause, T_RestClause, PreDict ).

b_t_modif( [++equal(L,R)|RestClause], [++equal(L,(NewSubterm->R))|T_RestClause], [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,_,S],
        cmd_arg( sortsnd, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,Ref,S],
	getnewvar( Ref ),
	b_t_modif( RestClause, T_RestClause, PreDict ),
        !.

b_t_modif( [++equal(L,R)|RestClause], [++equal(L,(NewSubterm->R))|T_RestClause], [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,S,_],
        cmd_arg( sortfst, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,S,Ref],
	getnewvar( Ref ),
	b_t_modif( RestClause, T_RestClause, PreDict ),
        !.

b_t_modif( [++equal(L,R)|RestClause], [++equal(L,(Ref->R))|T_RestClause], [(Funct,Ref)|PreDict]) :-
	!,
	R =.. [Funct|_],
	getnewvar( Ref ),
	b_t_modif( RestClause, T_RestClause, PreDict ).

b_t_modif( [--equal(L,R)|RestClause], [--equal(L,R)|T_RestClause], PreDict ) :-
	!,
	b_t_modif( RestClause, T_RestClause, PreDict ).

b_t_modif( [Literal|RestClause], [Literal|T_RestClause], PreDict ) :-
	b_t_modif( RestClause, T_RestClause, PreDict ).


%****************************************************************************
%* tp_modif( +Clause, -T_ClauseList, -T_PreDict )                           *
%****************************************************************************
%* - compile transitivity for positive equations                            *
%****************************************************************************

tp_modif( [], [], [] ).

tp_modif( [++equal(L,R)|RestClause], [++equal(L,R)|T_RestClause], PreDict ) :-
	ext_var( R ),
	!,
	tp_modif( RestClause, T_RestClause, PreDict ).

tp_modif( [++equal(L,R)|RestClause], [++equal(L,(NewSubterm->R))|T_RestClause], [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,_,S],
        cmd_arg( sortsnd, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,Ref,S],
	getnewvar( Ref ),
	tp_modif( RestClause, T_RestClause, PreDict ),
        !.

tp_modif( [++equal(L,R)|RestClause], [++equal(L,(NewSubterm->R))|T_RestClause], [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,S,_],
        cmd_arg( sortfst, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,S,Ref],
	getnewvar( Ref ),
	tp_modif( RestClause, T_RestClause, PreDict ),
        !.

tp_modif( [++equal(L,R)|RestClause], [++equal(L,(Ref->R))|T_RestClause],[(Funct,Ref)|PreDict]) :-
	!,
	R =.. [Funct|_],
	getnewvar( Ref ),
	tp_modif( RestClause, T_RestClause, PreDict ).

tp_modif( [Literal|RestClause], [Literal|T_RestClause], PreDict ) :-
	tp_modif( RestClause, T_RestClause, PreDict ).


%****************************************************************************
%* tpc_modif( +Clause, -T_ClauseList, -T_PreDict )                          *
%****************************************************************************
%* - compile transitivity for positive equations adding constraints         *
%****************************************************************************

tpc_modif( [], [], [] ).

tpc_modif( [++equal(L,R)|RestClause], [++equal(L,R)|T_RestClause], PreDict ) :-
        ext_var( R ),
        !,
	tpc_modif( RestClause, T_RestClause, PreDict ).

tpc_modif( [++equal(L,R)|RestClause], NewClause, [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,_,S],
        cmd_arg( sortsnd, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,Ref,S],
	getnewvar( Ref ),
        term_without_sorts( L, LWS ),
        term_without_sorts( R, RWS ),
        gen_ord_pred( (LWS,Ref), 0, OrdBipsLWS ),
        gen_ord_predeq( (RWS,Ref), 0, OrdBipsRWS ),
        append( OrdBipsLWS, OrdBipsRWS, OrdBips ),
        append( [++equal(L,(NewSubterm->R))|T_RestClause], OrdBips, NewClause ),
	tpc_modif( RestClause, T_RestClause, PreDict ),
        !.

tpc_modif( [++equal(L,R)|RestClause], NewClause, [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,S,_],
        cmd_arg( sortfst, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,S,Ref],
	getnewvar( Ref ),
        term_without_sorts( L, LWS ),
        term_without_sorts( R, RWS ),
        gen_ord_pred( (LWS,Ref), 0, OrdBipsLWS ),
        gen_ord_predeq( (RWS,Ref), 0, OrdBipsRWS ),
        append( OrdBipsLWS, OrdBipsRWS, OrdBips ),
        append( [++equal(L,(NewSubterm->R))|T_RestClause], OrdBips, NewClause ),
	tpc_modif( RestClause, T_RestClause, PreDict ),
        !.

tpc_modif( [++equal(L,R)|RestClause], NewClause,[(Funct,Ref)|PreDict]) :-
	!,
	R =.. [Funct|_],
	getnewvar( Ref ),
        term_without_sorts( L, LWS ),
        term_without_sorts( R, RWS ),
        gen_ord_pred( (LWS,Ref), 0, OrdBipsLWS ),
        gen_ord_predeq( (RWS,Ref), 0, OrdBipsRWS ),
        append( OrdBipsLWS, OrdBipsRWS, OrdBips ),
        append( [++equal(L,(Ref->R))|T_RestClause], OrdBips, NewClause ),
	tpc_modif( RestClause, T_RestClause, PreDict ).

tpc_modif( [Literal|RestClause], [Literal|T_RestClause], PreDict ) :-
	tpc_modif( RestClause, T_RestClause, PreDict ).


%****************************************************************************
%* tn_modif( +Clause, -T_ClauseList, -T_PreDict )                           *
%****************************************************************************
%* - compile transitivity for negative equations                            *
%****************************************************************************

tn_modif( [], [], [] ).

tn_modif( [--equal(L,R)|RestClause], [--equal(L,R)|T_RestClause], PreDict ) :-
	ext_var( R ),
	!,
	tn_modif( RestClause, T_RestClause, PreDict ).

tn_modif( [--equal(L,R)|RestClause], [--equal(L,(NewSubterm->R))|T_RestClause], [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,_,S],
        cmd_arg( sortsnd, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,Ref,S],
	getnewvar( Ref ),
	tn_modif( RestClause, T_RestClause, PreDict ),
        !.

tn_modif( [--equal(L,R)|RestClause], [--equal(L,(NewSubterm->R))|T_RestClause], [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,S,_],
        cmd_arg( sortfst, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,S,Ref],
	getnewvar( Ref ),
	tn_modif( RestClause, T_RestClause, PreDict ),
        !.

tn_modif( [--equal(L,R)|RestClause], [--equal(L,(Ref->R))|T_RestClause], [(Funct,Ref)|PreDict]) :-
	!,
	R =.. [Funct|_],
	getnewvar( Ref ),
	tn_modif( RestClause, T_RestClause, PreDict ).

tn_modif( [Literal|RestClause], [Literal|T_RestClause], PreDict ) :-
	tn_modif( RestClause, T_RestClause, PreDict ).


%****************************************************************************
%* tnc_modif( +Clause, -T_ClauseList, -T_PreDict )                          *
%****************************************************************************
%* - compile transitivity for negative equations adding constraints         *
%****************************************************************************

tnc_modif( [], [], [] ).

tnc_modif( [--equal(L,R)|RestClause], [--equal(L,R)|T_RestClause], PreDict ) :-
        ext_var( R ),
	!,
	tnc_modif( RestClause, T_RestClause, PreDict ).

tnc_modif( [--equal(L,R)|RestClause], NewClause, PreDict ) :-
        term_without_sorts( R, RWS ),
        var( RWS ),
	!,
        append( [--equal(L,R)], T_RestClause, NewClause ),
	tnc_modif( RestClause, T_RestClause, PreDict ).

tnc_modif( [--equal(L,R)|RestClause], NewClause, [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,_,S],
        cmd_arg( sortsnd, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,Ref,S],
	getnewvar( Ref ),
        term_without_sorts( L, LWS ),
        term_without_sorts( R, RWS ),
        gen_ord_predeq( (LWS,Ref), 0, OrdBipsLWS ),
        gen_ord_predeq( (RWS,Ref), 0, OrdBipsRWS ),
        append( OrdBipsLWS, OrdBipsRWS, OrdBips ),
        append( [--equal(L,(NewSubterm->R))|T_RestClause], OrdBips, NewClause ),
	tnc_modif( RestClause, T_RestClause, PreDict ),
        !.

tnc_modif( [--equal(L,R)|RestClause], NewClause, [(Funct,Ref)|PreDict]) :-
	R =.. [Funct,S,_],
        cmd_arg( sortfst, Sorts ),
        member( Funct, Sorts ),
        NewSubterm =.. [Funct,S,Ref],
	getnewvar( Ref ),
        term_without_sorts( L, LWS ),
        term_without_sorts( R, RWS ),
        gen_ord_predeq( (LWS,Ref), 0, OrdBipsLWS ),
        gen_ord_predeq( (RWS,Ref), 0, OrdBipsRWS ),
        append( OrdBipsLWS, OrdBipsRWS, OrdBips ),
        append( [--equal(L,(NewSubterm->R))|T_RestClause], OrdBips, NewClause ),
	tnc_modif( RestClause, T_RestClause, PreDict ),
        !.

tnc_modif( [--equal(L,R)|RestClause], NewClause, [(Funct,Ref)|PreDict]) :-
	!,
	R =.. [Funct|_],
	getnewvar( Ref ),
        term_without_sorts( L, LWS ),
        term_without_sorts( R, RWS ),
        gen_ord_predeq( (LWS,Ref), 0, OrdBipsLWS ),
        gen_ord_predeq( (RWS,Ref), 0, OrdBipsRWS ),
        append( OrdBipsLWS, OrdBipsRWS, OrdBips ),
        append( [--equal(L,(Ref->R))|T_RestClause], OrdBips, NewClause ),
	tnc_modif( RestClause, T_RestClause, PreDict ).

tnc_modif( [++equal(L,R)|RestClause], [++equal(L,R)|T_RestClause], PreDict ) :-
	!,
	tnc_modif( RestClause, T_RestClause, PreDict ).

tnc_modif( [Literal|RestClause], [Literal|T_RestClause], PreDict ) :-
	tnc_modif( RestClause, T_RestClause, PreDict ).


%****************************************************************************
%* e_modif( +Clause, -E_Clause, -PreDict )                                  *
%****************************************************************************
%* E-Modification: Flattening                                               *
%****************************************************************************

e_modif( LitList, ModLitList, PreDict ) :-
	e_modif_literal_list( LitList, ModLitList, PreDict ).


e_modif_literal_list( [], [], [] ).

e_modif_literal_list( [Lit|RestLiterals], ModLiterals, PreDict ) :-
	e_modif_literal( Lit, ModLitList, PreDictLit ),
	e_modif_literal_list( RestLiterals, ModRestLiterals, PreDictRest ),
	append( ModLitList, ModRestLiterals, ModLiterals ),
	append( PreDictLit, PreDictRest, PreDict ).


e_modif_literal( ++equal(L,R), [++equal(ModL,ModR)], PreDict ) :-
	!,
	e_modif_root_term( L, ModL, PreDictL ),
	e_modif_root_term( R, ModR, PreDictR ),
	append( PreDictL, PreDictR, PreDict ).

e_modif_literal( --equal(L,R), [], [] ) :-
	var(L), var(R),
	!,
	L=R.

e_modif_literal( --equal(L,R), [--equal(ModR,L)], PreDictR ) :-
	var(L), nonvar(R),
	!,
	e_modif_root_term( R, ModR, PreDictR ).

e_modif_literal( --equal(L,R), [--equal(ModL,ModR)], PreDict ) :-
	!,
	e_modif_root_term( L, ModL, PreDictL ),
	e_modif_term( R, ModR, PreDictR ),
	append( PreDictL, PreDictR, PreDict ).

e_modif_literal( ++Literal, [++ModLiteral], PreDict ) :-
	e_modif_root_term( Literal, ModLiteral, PreDict ).

e_modif_literal( --'$lpo'(L,R,Res), [--'$lpo'(L,R,Res)], [] ) :-
	!.

e_modif_literal( --'$lpoeq'(L,R,Res), [--'$lpoeq'(L,R,Res)], [] ) :-
	!.

e_modif_literal( --Literal, [--ModLiteral], PreDict ) :-
	e_modif_root_term( Literal, ModLiteral, PreDict ).


e_modif_root_term( Term, ModTerm, RealPreDict ) :-
	e_modif_term( Term, PreModTerm, PreDict ),
	( var(PreModTerm) ->
		( ModTerm=PreModTerm, RealPreDict=[] );
		( PreModTerm=(_->ModTerm), PreDict=[_|RealPreDict] )
	).

e_modif_term( Var, Var, [] ) :-
	var( Var ),
	!.

e_modif_term( Term, (Ref -> EModTerm), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Functor|Args],
	e_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ).


e_modif_term_list( [], [], [] ).

e_modif_term_list( [Var|RestArgs], [Var|EModRestArgs], PreDict ) :-
	var( Var ),
	!,
	e_modif_term_list( RestArgs, EModRestArgs, PreDict ).

e_modif_term_list( [Arg|RestArgs], [EModArg|EModRestArgs], PreDict ) :-
	e_modif_term( Arg, EModArg, PreDict1 ),
	e_modif_term_list( RestArgs, EModRestArgs, PreDictRest ),
	append( PreDict1, PreDictRest, PreDict ).


%****************************************************************************
%* b_e_modif( +Clause, -E_Clause, -PreDict )                                *
%****************************************************************************
%* Brand's E-Modification: Flattening                                       *
%****************************************************************************

b_e_modif( LitList, ModLitList, PreDict ) :-
	b_e_modif_literal_list( LitList, ModLitList, PreDict ).


b_e_modif_literal_list( [], [], [] ).

b_e_modif_literal_list( [Lit|RestLiterals], ModLiterals, PreDict ) :-
	b_e_modif_literal( Lit, ModLitList, PreDictLit ),
	b_e_modif_literal_list( RestLiterals, ModRestLiterals, PreDictRest ),
	append( ModLitList, ModRestLiterals, ModLiterals ),
	append( PreDictLit, PreDictRest, PreDict ).


b_e_modif_literal( ++equal(L,R), [++equal(ModL,ModR)], PreDict ) :-
	!,
	b_e_modif_root_term( L, ModL, PreDictL ),
	b_e_modif_root_term( R, ModR, PreDictR ),
	append( PreDictL, PreDictR, PreDict ).

b_e_modif_literal( --equal(L,R), [--equal(L,R)], [] ) :-
        ext_var(L), ext_var(R),
        !.

b_e_modif_literal( --equal(L,R), [--equal(ModL,ModR)], PreDict ) :-
	!,
	b_e_modif_root_term( L, ModL, PreDictL ),
	b_e_modif_root_term( R, ModR, PreDictR ),
	append( PreDictL, PreDictR, PreDict ).

b_e_modif_literal( ++Literal, [++ModLiteral], PreDict ) :-
	b_e_modif_root_term( Literal, ModLiteral, PreDict ).

b_e_modif_literal( --'$lpo'(L,R,Res), [--'$lpo'(L,R,Res)], [] ) :-
	!.

b_e_modif_literal( --'$lpoeq'(L,R,Res), [--'$lpoeq'(L,R,Res)], [] ) :-
	!.

b_e_modif_literal( --Literal, [--ModLiteral], PreDict ) :-
	b_e_modif_root_term( Literal, ModLiteral, PreDict ).


b_e_modif_root_term( Term, ModTerm, RealPreDict ) :-
	b_e_modif_term( Term, PreModTerm, PreDict ),
	( ext_var(PreModTerm) ->
		( ModTerm=PreModTerm, RealPreDict=[] );
		( PreModTerm=(_->ModTerm), PreDict=[_|RealPreDict] )
	).

b_e_modif_term( Var, Var, [] ) :-
	ext_var( Var ),
	!.

b_e_modif_term( Term, (NewLeft -> NewRight), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Funct,Subterm,S],
        cmd_arg( sortsnd, Sorts ),
        member( Funct, Sorts ),
        Subterm =.. [Functor|Args],
	b_e_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ),
        NewLeft =.. [Funct,Ref,S],
        NewRight =.. [Funct,EModTerm,S],
        !.

b_e_modif_term( Term, (NewLeft -> NewRight), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Funct,S,Subterm],
        cmd_arg( sortfst, Sorts ),
        member( Funct, Sorts ),
        Subterm =.. [Functor|Args],
	b_e_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ),
        NewLeft =.. [Funct,S,Ref],
        NewRight =.. [Funct,S,EModTerm],
        !.

b_e_modif_term( Term, (Ref -> EModTerm), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Functor|Args],
	b_e_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ).


b_e_modif_term_list( [], [], [] ).

b_e_modif_term_list( [Var|RestArgs], [Var|EModRestArgs], PreDict ) :-
	ext_var( Var ),
	!,
	b_e_modif_term_list( RestArgs, EModRestArgs, PreDict ).

b_e_modif_term_list( [Arg|RestArgs], [EModArg|EModRestArgs], PreDict ) :-
	b_e_modif_term( Arg, EModArg, PreDict1 ),
	b_e_modif_term_list( RestArgs, EModRestArgs, PreDictRest ),
	append( PreDict1, PreDictRest, PreDict).


%****************************************************************************
%* ml_modif( +Clause, -E_Clause, -PreDict )                                 *
%****************************************************************************
%* Flattening for left-hand sides                                           *
%****************************************************************************

ml_modif( LitList, ModLitList, PreDict ) :-
	ml_modif_literal_list( LitList, ModLitList, PreDict ).


ml_modif_literal_list( [], [], [] ).

ml_modif_literal_list( [Lit|RestLiterals], ModLiterals, PreDict ) :-
	ml_modif_literal( Lit, ModLitList, PreDictLit ),
	ml_modif_literal_list( RestLiterals, ModRestLiterals, PreDictRest ),
	append( ModLitList, ModRestLiterals, ModLiterals ),
	append( PreDictLit, PreDictRest, PreDict ).


ml_modif_literal( ++equal(L,R), [++equal(ModL,R)], PreDict ) :-
	!,
	ml_modif_root_term( L, ModL, PreDict ).

ml_modif_literal( --equal(L,R), [--equal(ModL,R)], PreDict ) :-
	!,
	ml_modif_root_term( L, ModL, PreDict ).

ml_modif_literal( ++Literal, [++ModLiteral], PreDict ) :-
	ml_modif_root_term( Literal, ModLiteral, PreDict ).

ml_modif_literal( --'$lpo'(L,R,Res), [--'$lpo'(L,R,Res)], [] ) :-
	!.

ml_modif_literal( --'$lpoeq'(L,R,Res), [--'$lpoeq'(L,R,Res)], [] ) :-
	!.

ml_modif_literal( --Literal, [--ModLiteral], PreDict ) :-
	ml_modif_root_term( Literal, ModLiteral, PreDict ).


ml_modif_root_term( Term, ModTerm, RealPreDict ) :-
	ml_modif_term( Term, PreModTerm, PreDict ),
	( ext_var(PreModTerm) ->
		( ModTerm=PreModTerm, RealPreDict=[] );
		( PreModTerm=(_->ModTerm), PreDict=[_|RealPreDict] )
	).

ml_modif_term( Var, Var, [] ) :-
	ext_var( Var ),
	!.

ml_modif_term( Term, (NewLeft -> NewRight), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Funct,Subterm,S],
        cmd_arg( sortsnd, Sorts ),
        member( Funct, Sorts ),
        Subterm =.. [Functor|Args],
	ml_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ),
        NewLeft =.. [Funct,Ref,S],
        NewRight =.. [Funct,EModTerm,S],
        !.

ml_modif_term( Term, (NewLeft -> NewRight), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Funct,S,Subterm],
        cmd_arg( sortfst, Sorts ),
        member( Funct, Sorts ),
        Subterm =.. [Functor|Args],
	ml_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ),
        NewLeft =.. [Funct,S,Ref],
        NewRight =.. [Funct,S,EModTerm],
        !.

ml_modif_term( Term, (Ref -> EModTerm), [(Functor,Ref)|PreDict] ) :-
	Term =.. [Functor|Args],
	ml_modif_term_list( Args, EModArgs, PreDict ),
	EModTerm =.. [Functor|EModArgs],
	getnewvar( Ref ).


ml_modif_term_list( [], [], [] ).

ml_modif_term_list( [Var|RestArgs], [Var|EModRestArgs], PreDict ) :-
	ext_var( Var ),
	!,
	ml_modif_term_list( RestArgs, EModRestArgs, PreDict ).

ml_modif_term_list( [Arg|RestArgs], [EModArg|EModRestArgs], PreDict ) :-
	ml_modif_term( Arg, EModArg, PreDict1 ),
	ml_modif_term_list( RestArgs, EModRestArgs, PreDictRest ),
	append( PreDict1, PreDictRest, PreDict ).


%****************************************************************************
%* predict2dict( +PreDict, +RelDict, -Dict )                                *
%****************************************************************************
%* convert a list of pairs (Functor,NewVar) to a dictionary wrt. RelDict    *
%****************************************************************************

predict2dict( [], _, [] ).

predict2dict( [(Funct,Ref)|PreDictRest], RelDict, 
					[UniqueArgSymb=Ref|NewVarDict] ) :-
	gen_unique_arg_symb( Funct, RelDict, 0, UniqueArgSymb ),
	predict2dict( PreDictRest, [UniqueArgSymb=Ref|RelDict], NewVarDict ).
	
gen_unique_arg_symb( ArgSymb, Dict, Number, UniqueArgSymb ) :-
	gen_corr_uc_symbol( ArgSymb, Number, NewArgSymb ),
	( my_member( (NewArgSymb=_), Dict ) ->
		( NextNumber is Number+1,
		  gen_unique_arg_symb( ArgSymb, Dict, NextNumber, UniqueArgSymb )
		);
		UniqueArgSymb = NewArgSymb
	).


%****************************************************************************
%* gen_companion_clause( +STE_Clause, +STE_Modified_CLause )                *
%****************************************************************************
%* create --equal subgoals for Var->Term links                              *
%****************************************************************************

gen_companion_clause( [], [] ).

gen_companion_clause( [Literal|RestLiterals], ModAllLiteralls ) :-
	gen_companion_lits( Literal, ModLiteral ),
	gen_companion_clause( RestLiterals, ModRestLiterals ),
	append( ModLiteral, ModRestLiterals, ModAllLiteralls ).


gen_companion_lits( ++Lit, [++ModLit|STE_Args] ) :-
	Lit =.. [Functor|Args],
	gen_companion_term_list( Args, ModArgs, STE_Args ),
	ModLit =.. [Functor|ModArgs].

gen_companion_lits( --Lit, [--ModLit|STE_Args] ) :-
	Lit =.. [Functor|Args],
	gen_companion_term_list( Args, ModArgs, STE_Args ),
	ModLit =.. [Functor|ModArgs].


gen_companion_term_list( [], [], [] ).

gen_companion_term_list( [Term|RestTerms], [ModTerm|RestModTerm], All_STE_Terms ) :-
	gen_companion_term( Term, ModTerm, STE_Terms ),
	gen_companion_term_list( RestTerms, RestModTerm, Rest_STE_Terms ),
	append( STE_Terms, Rest_STE_Terms, All_STE_Terms ).


gen_companion_term( Var, Var, [] ) :-
	ext_var( Var ),
	!.

gen_companion_term( (Var->Term), Var, NewSTE_Terms ) :-
	!,
	gen_companion_term( Term, ModTerm, STE_Terms ),
	append(STE_Terms,[--equal(ModTerm,Var)],NewSTE_Terms).

gen_companion_term( Term, ModTerm, STE_Terms ) :-
	Term =.. [Functor|Args],
	gen_companion_term_list( Args, ModArgs, STE_Terms ),
	ModTerm =.. [Functor|ModArgs].


%****************************************************************************
%* process_switch_stemod( +ArrangedSwitch )                                 *
%****************************************************************************
%* - what is to be done if -stemod is found                                 *
%*   this is the `original' moserstemod from CASC-14                        *
%****************************************************************************

process_switch_stemod( ["-stemod"] ) :-
	!,
	assert( cmd_arg(addrefleq,[]) ),
	assert( cmd_arg(smod,[]) ),
	assert( cmd_arg(tmod,[]) ),
	assert( cmd_arg(emod,[]) ),
	writeln( '# *** option: stemod' ).

process_switch_stemod( ["-stemod"|Arguments] ) :-
	( member('s',Arguments) ->
		assert( cmd_arg(smod,[]) )
		;
		true
	),
	( member('t',Arguments) ->
		( assert( cmd_arg(tmod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	( member('e',Arguments) ->
		( assert( cmd_arg(emod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	write( '# *** option: stemod(' ),
	write( Arguments ),
	writeln( ')' ).


%****************************************************************************
%* process_switch_brandstemod( +ArrangedSwitch )                            *
%****************************************************************************
%* - what is to be done if -brandstemod is found                            *
%****************************************************************************

process_switch_brandstemod( ["-brandstemod"] ) :-
	!,
	assert( cmd_arg(addrefleq,[]) ),
	assert( cmd_arg(smod,[]) ),
	assert( cmd_arg(b_tmod,[]) ),
	assert( cmd_arg(b_emod,[]) ),
	writeln( '# *** option: brandstemod' ).

process_switch_brandstemod( ["-brandstemod"|Arguments] ) :-
	( member('s',Arguments) ->
		assert( cmd_arg(smod,[]) )
		;
		true
	),
	( member('t',Arguments) ->
		( assert( cmd_arg(b_tmod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	( member('e',Arguments) ->
		( assert( cmd_arg(b_emod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	write( '# *** option: brandstemod(' ),
	write( Arguments ),
	writeln( ')' ).


%****************************************************************************
%* process_switch_moserstemod( +ArrangedSwitch )                            *
%****************************************************************************
%* - what is to be done if -moserstemod is found                            *
%****************************************************************************

process_switch_moserstemod( ["-moserstemod"] ) :-
	!,
	assert( cmd_arg(addrefleq,[]) ),
	assert( cmd_arg(smod,[]) ),
	assert( cmd_arg(tpmod,[]) ),
	assert( cmd_arg(remod,[]) ),
	assert( cmd_arg(tnmod,[]) ),
	assert( cmd_arg(mlmod,[]) ),
	writeln( '# *** option: moserstemod' ).

process_switch_moserstemod( ["-moserstemod"|Arguments] ) :-
	( member('s',Arguments) ->
		assert( cmd_arg(smod,[]) )
		;
		true
	),
	( member('tp',Arguments) ->
		( assert( cmd_arg(tpmod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	( member('re',Arguments) ->
		( assert( cmd_arg(remod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	( member('tn',Arguments) ->
		( assert( cmd_arg(tnmod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	( member('ml',Arguments) ->
		( assert( cmd_arg(mlmod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	write( '# *** option: moserstemod(' ),
	write( Arguments ),
	writeln( ')' ).


%****************************************************************************
%* process_switch_ceemod( +ArrangedSwitch )                                 *
%****************************************************************************
%* - what is to be done if -ceemod is found                                 *
%****************************************************************************

process_switch_ceemod( ["-ceemod"] ) :-
	!,
	assert( cmd_arg(addrefleq,[]) ),
	assert( cmd_arg(mmod,[]) ),
	assert( cmd_arg(remod,[]) ),
        assert( cmd_arg(smod,[]) ),
        assert( cmd_arg(snmod,[]) ),
	assert( cmd_arg(tpcmod,[]) ),
	assert( cmd_arg(tncmod,[]) ),
	writeln( '# *** option: ceemod' ).

process_switch_ceemod( ["-ceemod"|Arguments] ) :-
	( member('m',Arguments) ->
		( assert( cmd_arg(mmod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	( member('re',Arguments) ->
		( assert( cmd_arg(remod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	( member('s',Arguments) ->
		assert( cmd_arg(smod,[]) )
		;
		true
	),
	( member('sn',Arguments) ->
		assert( cmd_arg(snmod,[]) )
		;
		true
	),
	( member('tpc',Arguments) ->
		( assert( cmd_arg(tpcmod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	( member('tnc',Arguments) ->
		( assert( cmd_arg(tncmod,[]) ), assert( cmd_arg(addrefleq,[]) ) )
		;
		true
	),
	write( '# *** option: ceemod(' ),
	write( Arguments ),
	writeln( ')' ).


%****************************************************************************
%* test_switches_stemod                                                     *
%****************************************************************************
%* - check whether the original ste-modification really needs to be         *
%*   performed                                                              *
%****************************************************************************

test_switches_stemod :-
		% if formula contains no equality
	predaritylist( PredArityList ),
	\+ member( equal/_, PredArityList ),
	!,
		% then remove all equality modifications
	retract_all(cmd_arg(addrefleq,_)),
	retract_all(cmd_arg(smod,_)),
	retract_all(cmd_arg(tmod,_)),
	retract_all(cmd_arg(emod,_)).

test_switches_stemod :-
        modswitch(stemod),
		% if stemod was called but no rmeqax
	( cmd_arg(emod,_) ->
		L1 = [e] ; L1 = []
	),
	( cmd_arg(tmod,_) ->
		L2 = [t|L1] ; L2 = L1
	),
	( cmd_arg(smod,_) ->
		L3 = [s|L2] ; L3 = L2
	),
	( cmd_arg(addrefleq,_) ->
		L4 = [r|L3] ; L4 = L3
	),
	L4 \== [],
	\+ cmd_arg(rmeqax,_),
	!,
		% then remember to remove all equality axioms
	assert( cmd_arg(rmeqax,L4) ).

test_switches_stemod.


%****************************************************************************
%* test_switches_brandstemod                                                *
%****************************************************************************
%* - check whether Brand's ste-modification really needs to be performed    *
%****************************************************************************

test_switches_brandstemod :-
		% if formula contains no equality
	predaritylist( PredArityList ),
	\+ member( equal/_, PredArityList ),
	!,
		% then remove all equality modifications
	retract_all(cmd_arg(addrefleq,_)),
	retract_all(cmd_arg(smod,_)),
	retract_all(cmd_arg(b_tmod,_)),
	retract_all(cmd_arg(b_emod,_)).

test_switches_brandstemod :-
	modswitch(brandstemod),
	           % if brandstemod was called but no rmeqax
	( cmd_arg(b_emod,_) ->
		L1 = [e] ; L1 = []
	),
	( cmd_arg(b_tmod,_) ->
		L2 = [t|L1] ; L2 = L1
	),
	( cmd_arg(smod,_) ->
		L3 = [s|L2] ; L3 = L2
	),
	( cmd_arg(addrefleq,_) ->
		L4 = [r|L3] ; L4 = L3
	),
	L4 \== [],
	\+ cmd_arg(rmeqax,_),
	!,
		% then remember to remove all equality axioms
	assert( cmd_arg(rmeqax,L4) ).

test_switches_brandstemod.


%****************************************************************************
%* test_switches_moserstemod                                                *
%****************************************************************************
%* - check whether ste-modification really needs to be performed            *
%****************************************************************************

test_switches_moserstemod :-
		% if formula contains no equality
	predaritylist( PredArityList ),
	\+ member( equal/_, PredArityList ),
	!,
		% then remove all equality modifications
	retract_all(cmd_arg(addrefleq,_)),
	retract_all(cmd_arg(smod,_)),
	retract_all(cmd_arg(tpmod,_)),
	retract_all(cmd_arg(remod,_)),
	retract_all(cmd_arg(tnmod,_)),
	retract_all(cmd_arg(mlmod,_)).

test_switches_moserstemod :-
	modswitch(moserstemod),
		% if moserstemod was called but no rmeqax
	( cmd_arg(mlmod,_) ->
		L1 = [e] ; L1 = []
	),
	( cmd_arg(tnmod,_) ->
		L2 = [t|L1] ; L2 = L1
	),
	( cmd_arg(tpmod,_) ->
		L3 = [t|L2] ; L3 = L2
	),
	( cmd_arg(smod,_) ->
		L4 = [s|L3] ; L4 = L3
	),
	( cmd_arg(addrefleq,_) ->
		L5 = [r|L4] ; L5 = L4
	),
	L5 \== [],
	\+ cmd_arg(rmeqax,_),
	!,
		% then remember to remove all equality axioms
	assert( cmd_arg(rmeqax,L5) ).

test_switches_moserstemod.


%****************************************************************************
%* test_switches_ceemod                                                     *
%****************************************************************************
%* - check whether cee-modification really needs to be performed            *
%****************************************************************************

test_switches_ceemod :-
		% if formula contains no equality
	predaritylist( PredArityList ),
	\+ member( equal/_, PredArityList ),
	!,
		% then remove all equality modifications
	retract_all(cmd_arg(addrefleq,_)),
	retract_all(cmd_arg(mmod,_)),
	retract_all(cmd_arg(remod,_)),
        retract_all(cmd_arg(smod,_)),
        retract_all(cmd_arg(snmod,_)),
	retract_all(cmd_arg(tpcmod,_)),
	retract_all(cmd_arg(tncmod,_)).

test_switches_ceemod :-
	modswitch(ceemod),
		% if ceemod was called but no rmeqax
	( cmd_arg(tncmod,_) ->
		L1 = [t] ; L1 = []
	),
	( cmd_arg(tpcmod,_) ->
		L2 = [t|L1] ; L2 = L1
	),
	( cmd_arg(snmod,_) ->
		L3 = [s|L2] ; L3 = L2
	),
	( cmd_arg(smod,_) ->
		L4 = [s|L3] ; L4 = L3
	),
	( cmd_arg(remod,_) ->
		L5 = [r|L4] ; L5 = L4
	),
	( cmd_arg(mmod,_) ->
		L6 = [e|L5] ; L6 = L5
	),
	( cmd_arg(addrefleq,_) ->
		L7 = [r|L6] ; L7 = L6
	),
	L7 \== [],
	\+ cmd_arg(rmeqax,_),
	!,
		% then remember to remove all equality axioms
	assert( cmd_arg(rmeqax,L7) ).

test_switches_ceemod.

my_member( X, [Y] ) :-
        X == Y,
        !.

my_member( X, [_|Z] ) :-
        my_member( X, Z),
        !.
