%###########################################################################
%# mod_rmod.pl
%# - R modification, i.e. reflection of negative equations
%###########################################################################
%# predicates to be used from outside:
%#	- rmod/4
%#	- test_switches_rmod/0
%#	- process_switch_rmod/1
%# dependent on
%#	- mylib.pl
%###########################################################################
%#        Max Moser                         Automated Reasoning Research Group
%# Email: moser@informatik.tu-muenchen.de     Lehrstuhl VIII f"ur Informatik
%# Tel.:  +49-89/521096			     Technische Universit"at M"unchen
%# Fax.:  +49-89/526502			         D-80290 M"unchen, Germany
%# http://wwwjessen.informatik.tu-muenchen.de/~moser
%###########################################################################


%*****************************************************************************
%* rmod( clausewise, +Clause, +SwitchList, -NewClauseList )
%*****************************************************************************
%* - perform R modification
%*****************************************************************************

rmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify queries if required
        cmd_arg( noquerytrans, _ ),
	Clause = clause( _, conjecture, _, _, _ ),
	!.

rmod( clausewise, Clause, _, [Clause] ) :-
		% don't modify comments, rules and nomod...
	functor( Clause, Category, _ ),
	( Category=comment
	; Category=nomodclause
	; Category=rule
	; Category=nomodrule
	),
	!.

rmod( clausewise, Clause, [N], FinalClauseList ) :-
		% is N really a positive number?
	number( N ),
	N >= 0,
	!,
		% generate new clauses with N negative eqs
	gen_clauses_n_neg_eq( Clause, N, ClauseDelLitList ),
		% generate new variants for them
	gen_new_variants_list( ClauseDelLitList, NewVarClauseDelLitList ),
		% reflect deleted negative equations
	refl_deleted_neqs_list( NewVarClauseDelLitList, NewClauseList ),
		% finally add the original clause
	append( NewClauseList, [Clause], FinalClauseList ).

rmod( clausewise, Clause, [N], FinalClauseList ) :-
		% is N really a negative number?
	number( N ),
	N < 0,
	!,
	N1 is -N,
		% generate new clauses with up to N negative eqs
	gen_clauses_up_to_n_neg_eq( Clause, N1, ClauseDelLitList ),
		% generate new variants for them
	gen_new_variants_list( ClauseDelLitList, NewVarClauseDelLitList ),
		% reflect deleted negative equations
	refl_deleted_neqs_list( NewVarClauseDelLitList, NewClauseList ),
		% finally add the original clause
	append( NewClauseList, [Clause], FinalClauseList ).

%*****************************************************************************


%*****************************************************************************
%* gen_clauses_n_neg_eq( +Clause, +N, -ClauseDelLitList )
%*****************************************************************************
%* - generate a list of pairs where each pair consists of
%*   negative equations occuring in the literal list of the clause, and
%*   the original clause where the literal list is the same but the 
%*   deleted negative equations are missing and the clause contains
%*   N negative equations
%*****************************************************************************

gen_clauses_n_neg_eq( Clause, N, ClauseDelLitList ) :-
		% analyse Clause
	Clause = clause( Name, Type, LitList, Cons, Dict ),
		% now look for all possibilities to delete N neqs
	findall( (DelNeqs,clause(Name,Type,RestLitList,Cons,Dict)),
		delete_ordered_all_neqs_but_n( DelNeqs, LitList, 
				N, RestLitList ), ClauseDelLitList ).

%*****************************************************************************


%*****************************************************************************
%* gen_clauses_up_to_n_neg_eq( +Clause, +N, -ClauseDelLitList )
%*****************************************************************************
%* - generate a list of pairs where each pair consists of
%*   negative equations occuring in the literal list of the clause, and
%*   the original clause where the literal list is the same but the 
%*   deleted negative equations are missing and the clause contains
%*   up to N negative equations
%*****************************************************************************

gen_clauses_up_to_n_neg_eq( Clause, N, ClauseDelLitList ) :-
		% analyse Clause
	Clause = clause( Name, Type, LitList, Cons, Dict ),
		% now look for all possibilities to delete N neqs
	findall( (DelNeqs,clause(Name,Type,RestLitList,Cons,Dict)),
		delete_ordered_all_neqs_but_up_to_n( DelNeqs, LitList, 
				N, RestLitList ), ClauseDelLitList ).

%*****************************************************************************


%*****************************************************************************
%* delete_ordered_n_neqs( ?SubList, +LitList, +N, -RestLitList )
%*****************************************************************************
%* - delete list SubList of N neqs from FullList giving Rest
%* - the ordering of the elements in SubList is the same as in FullList
%*****************************************************************************

delete_ordered_n_neqs( [], RestAll, 0, RestAll ) :-
                % 0 neqs to delete so we are done
        !.

delete_ordered_n_neqs( [--equal(L,R)|RestDel], 
					[--equal(L,R)|RestAll], N, Rest ) :-
                % are still neqs to be deleted?
        N > 0,
        N1 is N-1,
                % continue to delete N-1 neqs from RestAll
        delete_ordered_n_neqs( RestDel, RestAll, N1, Rest ).

delete_ordered_n_neqs( RestDel, [Keep|RestAll], N, [Keep|Rest] ) :-
                % continue to delete N neqs from RestAll
        delete_ordered_n_neqs( RestDel, RestAll, N, Rest ).

%*****************************************************************************


%*****************************************************************************
%* delete_ordered_all_neqs_but_n( ?SubList, +LitList, +N, -RestLitList )
%*****************************************************************************
%* - delete list SubList of neqs from FullList giving Rest such that
%    Rest contains N neqs
%* - the ordering of the elements in SubList is the same as in FullList
%*****************************************************************************

delete_ordered_all_neqs_but_n( [], [], 0, [] ).

delete_ordered_all_neqs_but_n( RestDel, [--equal(L,R)|RestAll], N, 
						[--equal(L,R)|Rest] ) :-
                % keeping one neq
        N > 0,
        N1 is N-1,
                % continue to delete all but N-1 neqs from RestAll
        delete_ordered_all_neqs_but_n( RestDel, RestAll, N1, Rest ).

delete_ordered_all_neqs_but_n( RestDel, [SignedLit|RestAll], N, 
							[SignedLit|Rest] ) :-
		% just keep everything which is not a neq
	SignedLit \= --equal(_,_),
		% continue to delete all but N neqs from RestAll
        delete_ordered_all_neqs_but_n( RestDel, RestAll, N, Rest ).

delete_ordered_all_neqs_but_n( [--equal(L,R)|RestDel], [--equal(L,R)|RestAll], 
								N, Rest ) :-
		% continue to delete all but N neqs from RestAll
        delete_ordered_all_neqs_but_n( RestDel, RestAll, N, Rest ).

%*****************************************************************************


%*****************************************************************************
%* delete_ordered_all_neqs_but_up_to_n( ?SubList, +LitList, +N, -RestLitList )
%*****************************************************************************
%* - delete list SubList of neqs from FullList giving Rest such that
%    Rest contains N neqs
%* - the ordering of the elements in SubList is the same as in FullList
%*****************************************************************************

delete_ordered_all_neqs_but_up_to_n( [], [], _, [] ).

delete_ordered_all_neqs_but_up_to_n( RestDel, [--equal(L,R)|RestAll], N, 
						[--equal(L,R)|Rest] ) :-
                % keeping one neq
        N > 0,
        N1 is N-1,
                % continue to delete all but N-1 neqs from RestAll
        delete_ordered_all_neqs_but_up_to_n( RestDel, RestAll, N1, Rest ).

delete_ordered_all_neqs_but_up_to_n( RestDel, [SignedLit|RestAll], N, 
							[SignedLit|Rest] ) :-
		% just keep everything which is not a neq
	SignedLit \= --equal(_,_),
		% continue to delete all but N neqs from RestAll
        delete_ordered_all_neqs_but_up_to_n( RestDel, RestAll, N, Rest ).

delete_ordered_all_neqs_but_up_to_n( [--equal(L,R)|RestDel], 
					[--equal(L,R)|RestAll], N, Rest ) :-
		% continue to delete all but N neqs from RestAll
        delete_ordered_all_neqs_but_up_to_n( RestDel, RestAll, N, Rest ).

%*****************************************************************************


%*****************************************************************************
%* refl_deleted_neqs_list( +NewVarClauseDelLitList, -NewClauseList )
%*****************************************************************************
%* - ???
%*****************************************************************************

refl_deleted_neqs_list( [], [] ).

refl_deleted_neqs_list( [(DelNeqs,Clause)|Rest], [Clause|RestRefl] ) :-
		% try to reflect all deleted neqs
	refl_neq_list( DelNeqs ),
	!,
		% continue with other pairs
	refl_deleted_neqs_list( Rest, RestRefl ).

refl_deleted_neqs_list( [_|Rest], RestRefl ) :-
		% if not all neqs can be reflected then forget clause
	refl_deleted_neqs_list( Rest, RestRefl ).

%*****************************************************************************


%*****************************************************************************
%* refl_neq_list( ?NeqList )
%*****************************************************************************
%* - apply reflexivity to all neqs in NeqList
%*****************************************************************************

refl_neq_list( [] ).

refl_neq_list( [--equal(X,X)|Rest] ) :-
	refl_neq_list( Rest ).

%*****************************************************************************


%*****************************************************************************
%* process_switch_rmod( +ArrangedSwitch )
%*****************************************************************************
%* - what is to be done if -rmod is found
%*****************************************************************************

process_switch_rmod( ["-rmod",Count] ) :-
	cmd_arg( tptpout, _ ), !,
	assert( cmd_arg(rmod,[Count]) ),
	write( '% *** option: rmod(' ),
	write( Count ),
	writeln( ')' ).

process_switch_rmod( ["-rmod",Count] ) :-
	assert( cmd_arg(rmod,[Count]) ),
	write( '# *** option: rmod(' ),
	write( Count ),
	writeln( ')' ).

%*****************************************************************************


%*****************************************************************************
%* test_switches_rmod
%*****************************************************************************
%* - check whether r-modification really needs to be performed
%*****************************************************************************

test_switches_rmod :-
		% if formula contains no equality
	predaritylist( PredArityList ),
	\+ member( equal/_, PredArityList ),
	!,
		% then remove all equality modifications
	retract_all(cmd_arg(rmod,_)).

test_switches_rmod.

%*****************************************************************************
