%###########################################################################
%# mod_sig.pl
%# - augment clauses with signature information
%###########################################################################
%# predicates to be used from outside:
%#	- sig/4
%#	- process_switch_sig/1
%# dependent on
%#	- mylib.pl
%###########################################################################
%#        Max Moser                         Automated Reasoning Research Group
%# Email: moser@informatik.tu-muenchen.de     Lehrstuhl VIII f"ur Informatik
%# Tel.:  +49-89/521096			     Technische Universit"at M"unchen
%# Fax.:  +49-89/526502			         D-80290 M"unchen, Germany
%# http://wwwjessen.informatik.tu-muenchen.de/~moser
%###########################################################################


%*****************************************************************************
%* sig( problemwise, +ClauseList, +SwitchList, -NewClauseList )
%*****************************************************************************
%* - first produce signature information for all clauses
%* - then add declaration of predicate and function symbols
%* - then add signature info to clauses
%*****************************************************************************

sig( problemwise, ClauseList, SwitchList, [SigIntCl|NewClauseList] ) :-
		% test if global info should be added to certain clauses
	( member( c, SwitchList ) ; member( h, SwitchList ) ),
	!,
		% construct list of pairs of clauses and lists of
		% predicate and function symbols with arities
	build_clause_sig_list( ClauseList, ClauseSigList ),
		% generate clause to introduce signature
	gen_sig_intro_clause( SigIntCl ),
		% retrieve signature info for conjectures and hypothesises
	sel_sig_conj_hypo( ClauseSigList, FSigListConjHypo, PSigListConjHypo ),
		% add info to respective clauses
	append( FSigListConjHypo, PSigListConjHypo, SigListConjHypo ),
	add_clause_sig_to_clauselist( ClauseSigList, SwitchList, 
					SigListConjHypo, NewClauseList ).

sig( problemwise, ClauseList, _, [SigIntCl|NewClauseList] ) :-
		% construct list of pairs of clauses and predicate and
		% function symbols with arities
	build_clause_sig_list( ClauseList, ClauseSigList ),
		% generate clause to introduce signature
	gen_sig_intro_clause( SigIntCl ),
		% add info to respective clauses
	add_clause_sig_to_clauselist( ClauseSigList, [], [], NewClauseList ).

%*****************************************************************************


%*****************************************************************************
%* build_clause_sig_list( +ClauseList, -ClauseSigList )
%*****************************************************************************
%* - build up list of (Clause,Signature/Arity List of Clause)
%*****************************************************************************

build_clause_sig_list( [], [] ).
		% no clauses left, so we are done

build_clause_sig_list( [Clause|RestClauses], 
			[(Clause,FSig,PSig)|RestClauseSigList] ) :-
	Clause =.. [ ClauseCategory, _, _, LitList, _, _ ],
	( ClauseCategory = clause
	; ClauseCategory = rule
	; ClauseCategory = nomodclause
	; ClauseCategory = nomodrule
	),
	!,
		% get signature for current clause
	get_sig_for_clause( LitList, FSig, PSig ),
		% continue with rest clauses
	build_clause_sig_list( RestClauses, RestClauseSigList ).

build_clause_sig_list( [Clause|RestClauses], 
				[(Clause,_,_)|RestClauseSigList] ) :-
		% ignore comments, continue with rest clauses
	build_clause_sig_list( RestClauses, RestClauseSigList ).

%*****************************************************************************


%*****************************************************************************
%* get_sig_for_clause( +Clause, -FSig, -PSig )
%*****************************************************************************
%* - return list of function and predicate symbols with arrities
%*   occuring in Clause
%*****************************************************************************

get_sig_for_clause( LitList, FSig, PSig ) :-
		% forward recursion because of non-linearity
	get_sig_for_clause( LitList, [], [], DuplFSig, DuplPSig ),
		% remove duplicates
	sort( DuplFSig, FSig ),
	sort( DuplPSig, PSig ).


get_sig_for_clause( [], FSig, PSig, FSig, PSig ).

get_sig_for_clause( [LitOrTerm|RestLits], FSig, PSig, FullFSig, FullPSig ) :-
		% ignore variables
	var( LitOrTerm ),
	!,
		% so continue with remaining terms
	get_sig_for_clause( RestLits, FSig, PSig, FullFSig, FullPSig ).

get_sig_for_clause( [++Lit|RestLits], FSig, PSig, FullFSig, FullPSig ) :-
		% positive literal found
	!,
	Lit =.. [Symb|Args],
	functor( Lit, Symb, Arity ),
		% first process arguments
	get_sig_for_clause( Args, FSig, [Symb/Arity|PSig], NewFSig, NewPSig ),
		% then consider remaining literals
	get_sig_for_clause( RestLits, NewFSig, NewPSig, FullFSig, FullPSig ).

get_sig_for_clause( [--Lit|RestLits], FSig, PSig, FullFSig, FullPSig ) :-
		% ignore BIPs
	is_bip( Lit ),
	!,
	get_sig_for_clause( RestLits, FSig, PSig, FullFSig, FullPSig ).

get_sig_for_clause( [--Lit|RestLits], FSig, PSig, FullFSig, FullPSig ) :-
		% logical negative literal found
	!,
	Lit =.. [Symb|Args],
	functor( Lit, Symb, Arity ),
		% first process this literal
	get_sig_for_clause( Args, FSig, [Symb/Arity|PSig], NewFSig, NewPSig ),
		% then consider remaining literals
	get_sig_for_clause( RestLits, NewFSig, NewPSig, FullFSig, FullPSig ).

get_sig_for_clause( [Term|RestTerms], FSig, PSig, FullFSig, FullPSig ) :-
		% if above cases failed, then we have a term
	Term =.. [Symb|Args],
	functor( Term, Symb, Arity ),
		% first process its arguments
	get_sig_for_clause( Args, [Symb/Arity|FSig], PSig,  NewFSig, NewPSig ),
		% then consider remaining literals or terms
	get_sig_for_clause( RestTerms, NewFSig, NewPSig, FullFSig, FullPSig ).

%*****************************************************************************


%*****************************************************************************
%* sel_sig_conj_hypo( +ClauseSigList, -FSigListConjHypo, -PSigListConjHypo )
%*****************************************************************************
%* - get signature info of all conjecture and hypothesis clauses
%*****************************************************************************

sel_sig_conj_hypo( [], [], [] ).
		% all clauses processed so we are done

sel_sig_conj_hypo( [(Clause,FSig,PSig)|RestClauseSigList], NewFSig, NewPSig ) :-
		% analyse clause
	Clause =.. [ ClauseCategory, _, Type, _, _, _ ],
	( ClauseCategory = clause
	; ClauseCategory = rule
	; ClauseCategory = nomodclause
	; ClauseCategory = nomodrule
	),
		% is type=conjecture or hypothesis?
	( Type = conjecture ; Type = hypothesis ),
	!,
		% process remaining RestClauseSigList
	sel_sig_conj_hypo( RestClauseSigList, RestFSig, RestPSig ),
		% combine signatures
	append( FSig, RestFSig, CombinedFSig ),
	sort( CombinedFSig, NewFSig ),
	append( PSig, RestPSig, CombinedPSig ),
	sort( CombinedPSig, NewPSig ).

sel_sig_conj_hypo( [_|RestClauseSigList], NewFSig, NewPSig ) :-
		% consider RestClauseSigList
	sel_sig_conj_hypo( RestClauseSigList, NewFSig, NewPSig ).

%*****************************************************************************


%*****************************************************************************
%* add_clause_sig_to_clauselist( +ClauseSigList, +SwitchList, +CHSigList, 
%*							+NewClauseList )
%*****************************************************************************
%* - add signature information to all clauses in ClauseList
%*****************************************************************************

add_clause_sig_to_clauselist( [], _, _, [] ).
		% all clauses processed so we are done

add_clause_sig_to_clauselist( [ClauseSigTrip|RestClauses], SwitchList, 
				CHSigList, [NewClause|NewRestClauses] ) :-
		% process current clause
	add_clause_sig_to_clause( ClauseSigTrip, SwitchList, 
				CHSigList, NewClause ),
		% process rest
	add_clause_sig_to_clauselist( RestClauses, SwitchList,
					CHSigList, NewRestClauses ).

%*****************************************************************************


%*****************************************************************************
%* add_clause_sig_to_clause( +ClauseSigTrip, +SwitchList, +CHSigList, 
%*								-NewClause )
%*****************************************************************************
%* - if type of clause = conjecture and SwitchList contains 'c' 
%*   or type of clause = hypothesis and SwitchList contains 'h' 
%*   then add BIP with signature list
%* - else, just add BIP with local signature
%*****************************************************************************

add_clause_sig_to_clause( (Clause,_,_), SwitchList, CHSigList, NewClause ) :-
		% split up clause (no comments)
	Clause =.. [ ClauseCategory, Name, Type, LitList, Cons, Dict ],
	( ClauseCategory = clause
	; ClauseCategory = rule
	; ClauseCategory = nomodclause
	; ClauseCategory = nomodrule
	),
		% if SwitchList says that CHSigList is to be added
	( 
		( Type = conjecture, member( c, SwitchList ) )
		;
		( Type = hypothesis, member( h, SwitchList ) )
	),
	!,
		% get rid of information about arities
	funct_arity_list_2_funct_list( CHSigList, CHSymbList ),
		% construct new BIP
	SigLit =.. ['$sig',CHSymbList],
		% construct new clause
	NewClause =.. [ClauseCategory, Name, Type, [--SigLit|LitList], Cons, Dict].

add_clause_sig_to_clause( (Clause,FSig,PSig), _, _, NewClause ) :-
		% split up clause (no comments)
	Clause =.. [ ClauseCategory, Name, Type, LitList, Cons, Dict ],
	( ClauseCategory = clause
	; ClauseCategory = rule
	; ClauseCategory = nomodclause
	; ClauseCategory = nomodrule
	),
	!,
		% get rid of information about arities
	append( FSig, PSig, Sig ),
	funct_arity_list_2_funct_list( Sig, Fun ),
		% construct new BIP
	SigLit =.. ['$sig',Fun],
		% construct new clause
	NewClause =.. [ClauseCategory, Name, Type, [--SigLit|LitList], Cons, Dict].

add_clause_sig_to_clause( (Clause,_,_), _, _, Clause ).
		% don't modify other clauses, e.g. comments

%*****************************************************************************


%*****************************************************************************
%* funct_arity_list_2_anon_funct_list( +FuncArityList, -OpList )
%*****************************************************************************
%* generate list of symbols with `_' arguments
%*****************************************************************************

funct_arity_list_2_anon_funct_list( [], [] ).
		% all symbols processed

funct_arity_list_2_anon_funct_list( [F/Ar|Rest], [AnonTerm|RestOpList] ) :-
		% get Ar anonymous variables
	get_list_of_n( Ar, '_', ListOfUnderscores ),
		% build up new anonymous term
	AnonTerm =.. [F|ListOfUnderscores],
		% continue with rest
	funct_arity_list_2_anon_funct_list( Rest, RestOpList ).

%*****************************************************************************


%*****************************************************************************
%* funct_arity_list_2_funct_list( +FuncArityList, -OpList )
%*****************************************************************************
%* generate list of symbols 
%*****************************************************************************

funct_arity_list_2_funct_list( [], [] ).
		% all symbols processed

funct_arity_list_2_funct_list( [F/_|Rest], [F|RestOpList] ) :-
		% continue with rest
	funct_arity_list_2_funct_list( Rest, RestOpList ).

%*****************************************************************************


%*****************************************************************************
%* gen_sig_intro_clause( -SigIntroClause )
%*****************************************************************************
%* - generate a new clause which only serves to introduce the signature
%*   of the whole formula
%*****************************************************************************

gen_sig_intro_clause( SigIntCl ) :-
		% collect signature info for whole formula
	predaritylist( PSig ),
	functaritylist( FSig ),
		% generate anonymous terms
	funct_arity_list_2_anon_funct_list( FSig, OpList ),
	funct_arity_list_2_anon_funct_list( PSig, NewLitList ),
		% add -- to literals in NewLitList
	add_neg_sign_to_lit_list( NewLitList, SignedNewLitList ),
		% build up new clause
	SigIntCl = nomodclause( signature_list_of_whole_formula, axiom,
		[++signature_list_of_whole_formula(OpList)|SignedNewLitList],
			[], [] ).

%*****************************************************************************


%*****************************************************************************
%* add_neg_sign_to_lit_list( +LitList, -NegLitList )
%*****************************************************************************
%* - add -- to all literals in LitList
%*****************************************************************************

add_neg_sign_to_lit_list( [], [] ).

add_neg_sign_to_lit_list( [Lit|Rest], [--Lit|NegRest] ) :-
	add_neg_sign_to_lit_list( Rest, NegRest ).

%*****************************************************************************


%*****************************************************************************
%* process_switch_sig( +ArrangedSwitch )
%*****************************************************************************
%* - what is to be done if -sig is found
%*****************************************************************************

process_switch_sig( ["-sig"|Arguments] ) :-
	assert( cmd_arg(sig,Arguments) ),
	write( '# *** option: sig(' ),
	write( Arguments ),
	writeln( ')' ).

%*****************************************************************************

