%###########################################################################
%# rmeqax.pl
%# - remove equality axioms
%###########################################################################
%# predicates to be used from outside:
%#	- rmeqax/4
%#	- process_switch_rmeqax/1
%#	- test_switches_rmeqax/0
%# dependent on
%#	- clause info facts generated by siginfo.pl
%###########################################################################
%#        Max Moser                         Automated Reasoning Research Group
%# Email: moser@informatik.tu-muenchen.de     Lehrstuhl VIII f"ur Informatik
%# Tel.:  +49-89/521096			     Technische Universit"at M"unchen
%# Fax.:  +49-89/526502			         D-80290 M"unchen, Germany
%# http://wwwjessen.informatik.tu-muenchen.de/~moser
%###########################################################################


%*****************************************************************************
%* rmeqax( clausewise, +Clause, +SwitchList, -NewClauseList )
%*****************************************************************************
%* - remove undesired equality axioms
%*****************************************************************************

rmeqax( clausewise, Clause, SwitchList, [comment(Comment)] ) :-
		% only modify clauses
	(
		Clause = clause( Name, Type, LitList, _, _ ) ;
		Clause = rule( Name, Type, LitList, _, _ )
	),
		% if switch list is empty, set switchlist 
	( SwitchList == [] 
		->	CurrentSwitchList = [r,s,t,e] ;
			CurrentSwitchList = SwitchList
	),
		% test whether clause is an unwanted clause
	unwanted_equality_axiom( LitList, CurrentSwitchList ),
	!,
		% construct comment saying that clause has been deleted
	name( 'removed: ', RemovedChars ),
	name( Name, NameChars ),
	name( ', ', CommaChars ),
	name( Type, TypeChars ),
	append( CommaChars, TypeChars, TmpChars1 ),
	append( NameChars, TmpChars1, TmpChars2 ),
	append( RemovedChars, TmpChars2, TmpChars3 ),
	name( Comment, TmpChars3 ).

rmeqax( clausewise, Clause, _, [Clause] ).
		% everything else does not need to be modified

%*****************************************************************************


%*****************************************************************************
%* unwanted_equality_axiom( +LitList, +Mods )
%*****************************************************************************
%* - check if the literals form an unwanted clause
%* - mods can be r,s,t,e
%* - no need to test for variables because instances are subsumed
%*****************************************************************************

unwanted_equality_axiom( LitList, Mods ):-
		% remove reflexivity?
	member( r, Mods ),
		% clause is subsumed by reflexivity?
	member( ++equal(X1,X2), LitList ),
	X1 == X2,
	!.

unwanted_equality_axiom( Literals, Mods ):-
		% remove symmetry?
	member( s, Mods ),
		% clause is subsumed by symmetry?
	delete( ++equal(X1,Y1), Literals, RestLiterals ),
	member( --equal(Y2,X2), RestLiterals ),
	X1 == X2,
	Y1 == Y2,
	!.

unwanted_equality_axiom( Literals, Mods ):-
		% remove transitivity?
	member( t, Mods ),
		% clause is subsumed by transitivity?
	delete( ++equal(X1,Z), Literals, RestLiterals1 ),
	delete( --equal(X2,Y), RestLiterals1, RestLiterals2 ),
	X1 == X2,
	eq_lits_for_var_pairs( [Z], [Y], RestLiterals2 ),
	!.

unwanted_equality_axiom( Literals, Mods ):-
		% remove substitutivity?
	member( e, Mods ),
		% clause is subsumed by functional substitutivity?
	delete( ++equal(X,Y), Literals, RestLiterals ),
	nonvar( X ),
	nonvar( Y ),
	X =.. [Fun|XArgs],
	Y =.. [Fun|YArgs],
	eq_lits_for_var_pairs( XArgs, YArgs, RestLiterals ),
	!.

unwanted_equality_axiom( Literals, Mods ):-
		% remove substitutivity?
	member( e, Mods ),
		% clause is subsumed by predicative substitutivity?
	delete( ++LitP, Literals, RestLiterals1 ),
	LitP =.. [Pred|PArgs],
	Pred \== equal,
	delete( --LitN, RestLiterals1, RestLiterals2 ),
	LitN =.. [Pred|NArgs],
	eq_lits_for_var_pairs( PArgs, NArgs, RestLiterals2 ),
	!.

%*****************************************************************************


%*****************************************************************************
%* eq_lits_for_var_pairs( +VarList1, +VarList2, +LitList )
%*****************************************************************************
%* - check whether for every corresponding pair of variables from both lists
%*   there is a literal --equal(Var1,Var2) in LitList
%* - LitList may contain additional literals, no problem
%* - no need to test whether we have indeed variables
%*****************************************************************************

eq_lits_for_var_pairs( [], [], _ ).
		% forget about additional literals

eq_lits_for_var_pairs( [Var1|VarList1], [Var2|VarList2], LitList ) :-
		% if both variables are identical then forget LitList
	Var1 == Var2,
	!,
		% continue search
	eq_lits_for_var_pairs( VarList1, VarList2, LitList ).

eq_lits_for_var_pairs( [Var1|VarList1], [Var2|VarList2], LitList ) :-
		% look for equality literal for different variables
	delete( --equal(V1,V2), LitList, RestLitList ),
		% test whether variables are the same
	( (V1==Var1,V2==Var2) ; (V1==Var2,V2==Var1) ),
		% continue search
	eq_lits_for_var_pairs( VarList1, VarList2, RestLitList ).

%*****************************************************************************


%*****************************************************************************
%* process_switch_rmeqax( +ArrangedSwitch )
%*****************************************************************************
%* - what is to be done if -rmeqax is found
%*****************************************************************************

process_switch_rmeqax( ["-rmeqax"|Arguments] ) :-
	cmd_arg( tptpout, _ ), !,
	assert( cmd_arg(rmeqax,Arguments) ),
	write( '% *** option: rmeqax(' ),
	write( Arguments ),
	writeln( ')' ).

process_switch_rmeqax( ["-rmeqax"|Arguments] ) :-
	assert( cmd_arg(rmeqax,Arguments) ),
	write( '# *** option: rmeqax(' ),
	write( Arguments ),
	writeln( ')' ).

%*****************************************************************************


%*****************************************************************************
%* test_switches_rmeqax
%*****************************************************************************
%* - check whether deletion of equality axioms really needs to be performed
%*****************************************************************************

test_switches_rmeqax :-
		% if formula contains no equality
	predaritylist( PredArityList ),
	\+ member( equal/_, PredArityList ),
	!,
		% then remove rmeqax
	retract_all(cmd_arg(rmeqax,_)).

test_switches_rmeqax.

%*****************************************************************************
